"""Handles pushing data to the aggregator"""

import sys
import zipfile
from pathlib import Path

import requests
import rich
from pandas import read_parquet

from cumulus_library import base_utils, const


def upload_data(
    progress_bar: rich.progress.Progress,
    file_upload_progress: rich.progress.TaskID,
    file_path: Path,
    version: str,
    args: dict,
) -> str:
    """Fetches presigned url and uploads file to aggregator"""
    study = file_path.parts[-2]
    file_name = file_path.parts[-1]
    c = rich.get_console()
    progress_bar.update(file_upload_progress, description=f"Uploading {study}/{file_name}")
    url = args["url"]
    if args["network"]:
        # coercion to handle optional presence of trailing slash in the url
        url = url.rstrip("/") + "/" + args["network"]
    prefetch_res = requests.post(
        url,
        json={
            "study": study,
            "data_package_version": int(float(version)),
            "filename": f"{args['user']}_{file_name}",
        },
        auth=(args["user"], args["id"]),
        timeout=60,
    )
    if args["preview"]:
        c.print("prefetch request")
        c.print("headers", prefetch_res.request.headers)
        c.print("body", prefetch_res.request.body, "\n")
        c.print("response")
        c.print(prefetch_res.json(), "\n")
    if prefetch_res.status_code == 412:
        sys.exit(str(prefetch_res.json()))
    elif prefetch_res.status_code != 200:
        c.print("Invalid user/site id")
        prefetch_res.raise_for_status()
    transaction_id = prefetch_res.headers.get("transaction-id")
    res_body = prefetch_res.json()
    with open(file_path, "rb") as data_file:
        files = {"file": (file_name, data_file)}
        upload_req = requests.Request(
            "POST", res_body["url"], data=res_body["fields"], files=files
        ).prepare()
        if not args["preview"]:
            s = requests.Session()
            upload_res = s.send(upload_req, timeout=60)
            if upload_res.status_code != 204:
                c.print(f"Error uploading {study}/{file_name}")
                upload_res.raise_for_status()
        else:
            c.print("upload_req")
            c.print("headers", upload_req.headers)
            c.print("body", upload_req.body, "\n")
    progress_bar.update(file_upload_progress, advance=1)
    return transaction_id


def upload_files(args: dict):
    """Wrapper to prep files & console output"""
    if args["data_path"] is None:
        sys.exit("No data directory provided - please provide a path to your study export folder.")
    file_paths = list(args["data_path"].glob("**/*.zip"))
    filtered_paths = []
    if not args["user"] or not args["id"]:
        sys.exit("user/id not provided, please pass --user and --id")
    for target in args["target"]:
        for path in file_paths:
            if path.parent.name == target and path.name == f"{target}.zip":
                filtered_paths.append(path)
        if len(filtered_paths) == 0:
            sys.exit("No files found for upload. Is your data path/target specified correctly?")
        archive_path = filtered_paths[0]
        upload_archive = zipfile.ZipFile(archive_path)
        archive_contents = upload_archive.namelist()
        invalid_contents = []
        for file in archive_contents:
            if not any(x in file for x in const.ALLOWED_UPLOADS):
                invalid_contents.append(file)
        if len(invalid_contents) > 0:
            sys.exit(
                f"{archive_path} contains files that are not allowed:"
                f"  {invalid_contents}"
                "This likely means you tried to upload an archive containing line level data, "
                "but may also be a bug related to your study export names."
            )
        if target != "discovery":
            if not any(f"{target}__meta_date.meta.parquet" == x for x in archive_contents):
                sys.exit(
                    f"Study '{target}' does not contain a {target}__meta_date table.\n"
                    "See the documentation for more information about this required table.\n"
                    "https://docs.smarthealthit.org/cumulus/library/creating-studies.html#metadata-tables"
                )
        try:
            meta_version = next(
                filter(lambda x: str(x).endswith("__meta_version.meta.parquet"), archive_contents)
            )
            version = str(
                read_parquet(upload_archive.open(meta_version))["data_package_version"][0]
            )
        except StopIteration:
            version = "0"
        # TODO: I looked into monitoring upload progress instead of completed files and it is
        # non-trivial - potential point for improvement later
        with base_utils.get_progress_bar() as progress_bar:
            file_upload_progress = progress_bar.add_task(f"Uploading {target}...", total=1)
            upload_data(progress_bar, file_upload_progress, filtered_paths[0], version, args)



"""Module for generating core medicationrequest table"""

import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "medicationrequest": {
        "id": [],
        "status": [],
        "intent": [],
        "authoredOn": [],
        "requester": sql_utils.REFERENCE,
        "reportedBoolean": [],
        "reportedReference": sql_utils.REFERENCE,
        "subject": sql_utils.REFERENCE,
        "encounter": sql_utils.REFERENCE,
        "medicationReference": sql_utils.REFERENCE,
    }
}


class MedicationRequestBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating MedicationRequest table..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ) -> None:
        """Constructs queries related to medication requests

        :param config: A study config object
        """
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries += [
            core_templates.get_core_template("medicationrequest", validated_schema),
        ]



"""Database abstraction layer for DuckDB

See base.py for design rules of thumb.

Since duckdb tends to be more malleable than a cloud-based database, if you
need to paper over differences, it's best to use the connection.create_function
pattern here to get sql flavors in alignment.
"""

import base64
import collections
import datetime
import json
import pathlib
import re
import time
from concurrent import futures

import duckdb
import pandas
import pyarrow
from rich import progress

from cumulus_library import base_utils
from cumulus_library.databases import base, utils


class DuckDatabaseBackend(base.DatabaseBackend):
    """Database backend that uses local files via duckdb"""

    def __init__(
        self,
        db_file: str,
        schema_name: str | None = None,
        max_concurrent: int | None = None,
        pyarrow_cache_path: str | None = None,
    ):
        super().__init__("main")
        self.db_type = "duckdb"
        self.db_file = db_file
        self.connection = None
        self.max_concurrent = max_concurrent or 20
        self.pyarrow_cache_path = pyarrow_cache_path

    def init_errors(self):
        return ["Binder Error", "Catalog Error"]

    def connect(self):
        """Connects to the local duckdb database"""
        # As of the 1.0 duckdb release, local scopes, where schema names can be provided
        # as configuration to duckdb.connect, are not supported.
        # https://duckdb.org/docs/sql/statements/set.html#syntax
        # This is where the connection config would be supplied when it is supported
        self.connection = duckdb.connect(self.db_file)
        # Aliasing Athena's as_pandas to duckDB's df cast
        duckdb.DuckDBPyConnection.as_pandas = duckdb.DuckDBPyConnection.df

        # Ignore order of NDJSON that we load in. It saves memory and order doesn't matter for us.
        self.connection.execute("SET preserve_insertion_order = false;")

        # Are we getting a cached copy of the pyarrow schema when the class was created?
        # Usually this means we're running unit tests and trying to save time
        if self.pyarrow_cache_path and pathlib.Path(self.pyarrow_cache_path).is_file():
            self.connection.execute(
                "CREATE TABLE IF NOT EXISTS pyarrow_cache AS "  # noqa: S608
                f"SELECT * FROM read_parquet('{self.pyarrow_cache_path}')"
            )

        # Paper over some syntax differences between Athena and DuckDB
        self.connection.create_function(
            # DuckDB's version is array_to_string -- seems there is no standard here.
            "array_join",
            self._compat_array_join,
            None,
            duckdb.sqltypes.VARCHAR,
        )
        self.connection.create_function(
            # DuckDB's version is regexp_matches.
            "regexp_like",
            self._compat_regexp_like,
            None,
            duckdb.sqltypes.BOOLEAN,
        )
        self.connection.create_function(
            "from_iso8601_timestamp",
            self._compat_from_iso8601_timestamp,
            None,
            # Note: DuckDB provides a timestamp aware column type, TIMESTAMP_TZ, but
            # as of this writing on version 1.4.1, it is doing some casting to local
            # offset time rather than timezone, which we're electing to not deal with
            duckdb.sqltypes.TIMESTAMP,
        )
        self.connection.create_function(
            # When trying to calculate an MD5 hash in Trino/Athena, the implementation
            # expects to recieve a varbinary type, so if you're hashing a string,
            # you would invoke it like `SELECT md5(to_utf8(string_col)) from table`.
            #
            # DuckDB's md5() function accepts a varchar instead, and does not have a
            # to_utf() function or varbinary type, so we patch this with a UDF that
            # just provides back the original string. As a result, these functions
            # have different signatures, but for cases like this where you're
            # conforming an argument to another function, it provides appropriate
            # function mocking
            #
            # NOTE: currently we do not have a use case beyond experimentation where
            # using MD5 hashes provide a benefit. Until we do, it is not required to
            # support this in other DatabaseBackend implementations.
            "to_utf8",
            self._compat_to_utf8,
            None,
            duckdb.sqltypes.VARCHAR,
        )

    def insert_tables(self, tables: dict[dict[str, str]]) -> None:
        """Ingests all ndjson data from a folder tree.

        This function will write a cache of pyarrow datasets to the database.
        The next time this database is started, if we get passed the same set
        of files and they haven't changed, we'll use the cache. Otherwise
        we'll make a new set.

        The data loaded in this way is often from the output folder of Cumulus ETL

        :param tables: A dict describing table info (generated by utils.get_ndjson_files())"""

        self.connection.execute(
            "CREATE TABLE IF NOT EXISTS pyarrow_cache "
            "(table_name VARCHAR, file_array VARCHAR, schema VARCHAR)"
        )

        cache_valid = self.is_cache_valid(tables)

        if not cache_valid:
            # Our cache is out of date, let's get fresh table datasets - this may take a bit,
            # since read_ndjson_dir has to scan through all the files.
            datasets = utils.read_ndjson_dir(None, tables)
            self.connection.execute("DELETE FROM pyarrow_cache")
        else:
            # Our cache is still good. We'll pipe back in the pyarrow schemas we serialized
            # out to the database, and use that to create the dataset interfaces.
            datasets = self.get_cached_datasets()

        for name, dataset in datasets.items():
            self.connection.register(f"{name}", dataset)

            # If we've got new datasets, let's stash information about the expected files
            # and the schemas we found in the db for the next run.
            if not cache_valid:
                self.cache_dataset(name, dataset)

    def is_cache_valid(self, tables) -> bool:
        """Checks if ndjson has been updated since the last time we cached it"""
        # 'tables' contains a group of fhir tables, as well as etl completion tables;
        # we don't care about that distinction here, so we'll just ignore the first
        # layer of keys.
        for subtables in tables.values():
            for name, fragments in subtables.items():
                # See if we've got a cache of this table, and if it's still valid
                cached_files = self.connection.execute(
                    f"SELECT file_array FROM pyarrow_cache WHERE table_name = '{name.lower()}'"  # noqa: S608
                ).fetchone()
                if cached_files is None:
                    return False  # there is no cache for this table, so don't try to load it
                cached_files = json.loads(cached_files[0])

                fragment_infos = []
                for fragment in fragments:
                    fragment_modtime = pathlib.Path(fragment).lstat().st_mtime
                    fragment_infos.append([fragment, fragment_modtime])
                if cached_files != fragment_infos:
                    return False
        return True

    def cache_dataset(self, name: str, dataset: pyarrow.dataset.Dataset):
        """serializes a pyarrow Dataset and stashes it in the cache table"""
        fragment_infos = []
        fragments = dataset.get_fragments()
        for fragment in fragments:
            fragment_path = fragment.path
            fragment_modtime = pathlib.Path(fragment_path).lstat().st_mtime
            fragment_infos.append((fragment_path, fragment_modtime))
        schema = base64.b64encode(dataset.schema.serialize().to_pybytes()).decode("ascii")
        self.connection.execute(
            "INSERT INTO pyarrow_cache "  # noqa: S608
            f"VALUES ('{name}', '{json.dumps(fragment_infos)}', '{schema}')"
        )

    def get_cached_datasets(self):
        """Deserializes pyarrow datasets from a database cache"""
        datasets = {}
        schemas = self.connection.execute(
            "SELECT table_name, file_array, schema FROM pyarrow_cache"
        ).fetchall()
        for row in schemas:
            schema = pyarrow.ipc.read_schema(pyarrow.py_buffer(base64.b64decode(row[2])))
            files = []
            for source in json.loads(row[1]):
                files.append(source[0])
            datasets[row[0]] = pyarrow.dataset.dataset(
                files, schema=schema, format=utils._json_format()
            )
        return datasets

    @staticmethod
    def _compat_array_join(value: list[str | None], delimiter: str | None) -> str:
        if delimiter is None or delimiter == "None":
            # This is exercised locally on unit tests but is not in CI. Not sure why,
            # and not sure it's worth debugging
            delimiter = ""  # pragma: no cover
        return delimiter.join(v for v in value if v is not None)

    @staticmethod
    def _compat_regexp_like(string: str | None, pattern: str | None) -> bool:
        match = re.search(pattern, string)
        return match is not None

    @staticmethod
    def _compat_to_utf8(value: str | None) -> datetime.date | None:
        """See the create_function() call for to_utf8 for more background"""
        # This is exercised locally on unit tests but is not in CI. Not sure why,
        # and not sure it's worth debugging
        return value  # pragma: no cover

    @staticmethod
    def _compat_from_iso8601_timestamp(
        value: str,
    ) -> datetime.datetime:
        # handle partial dates like 1970 or 1980-12 (which spec allows)
        if len(value) < 10:
            pieces = value.split("-")
            if len(pieces) == 1:
                return datetime.datetime(int(pieces[0]), 1, 1, tzinfo=datetime.UTC)
            else:
                return datetime.datetime(int(pieces[0]), int(pieces[1]), 1, tzinfo=datetime.UTC)

        dt = datetime.datetime.fromisoformat(value)
        if not dt.tzinfo:
            return dt.replace(tzinfo=datetime.UTC)
        return dt.astimezone(datetime.UTC)

    def cursor(self) -> duckdb.DuckDBPyConnection:
        return self.connection

    def pandas_cursor(self) -> duckdb.DuckDBPyConnection:
        # Since this is not provided, return the vanilla cursor
        return self.connection

    def execute_as_pandas(
        self, sql: str, chunksize: int | None = None
    ) -> (pandas.DataFrame | collections.abc.Iterator[pandas.DataFrame], list[tuple]):
        # We call convert_dtypes here in case there are integer columns.
        # Pandas will normally cast nullable-int as a float type unless
        # we call this to convert to its nullable int column type.
        # PyAthena seems to do this correctly for us, but not DuckDB.
        result = self.connection.execute(sql)
        if chunksize:
            return iter([result.df().convert_dtypes()]), result.description
        return result.df().convert_dtypes(), result.description

    def parser(self) -> base.DatabaseParser:
        return DuckDbParser()

    def operational_errors(self) -> tuple[type[Exception], ...]:
        return (
            duckdb.OperationalError,
            duckdb.BinderException,
        )

    def export_table_as_parquet(
        self, table_name: str, file_name: str, location: pathlib.Path, *args, **kwargs
    ) -> bool:
        parquet_path = location / f"{file_name}"
        table_size = self.connection.execute(f"SELECT count(*) FROM {table_name}").fetchone()  # noqa: S608
        if table_size[0] == 0:
            return False
        query = f"""COPY
            (SELECT * FROM {table_name})
            TO '{parquet_path}'
            (FORMAT parquet)
            """  # noqa: S608
        self.connection.execute(query)
        return True

    def _write_thread(self, query, verbose, progress_bar, task, query_console_output, datasets):
        thread_con = self.connection.cursor()
        # Since registrations are per cursor, we'll use our cache
        # of pyarrow datasets again to re-initialize ndjson tables
        for name, dataset in datasets.items():
            thread_con.register(f"{name}", dataset)
        with query_console_output(verbose, query, progress_bar, task):
            thread_con.execute(query)

    def parallel_write(
        self,
        queries: list[str],
        verbose: bool,
        progress_bar: progress.Progress,
        task: progress.Task,
    ):
        datasets = self.get_cached_datasets()
        with futures.ThreadPoolExecutor(max_workers=self.max_concurrent) as executor:
            res = []
            for query in queries:
                res.append(
                    (
                        query,
                        executor.submit(
                            self._write_thread,
                            query,
                            verbose,
                            progress_bar,
                            task,
                            base_utils.query_console_output,
                            datasets,
                        ),
                    )
                )
            utils.handle_concurrent_errors(res, self.db_type)

        # This is a temporary workaround for cases where duckdb is still doing :something:
        # after the threads have resolved, where if you try to recreate a table it may throw
        # an error, which is only affecting the --statistics flag ¯\_(ツ)_/¯
        time.sleep(0.25)

    def create_schema(self, schema_name):
        """Creates a new schema object inside the database"""
        schemas = self.connection.sql(
            "SELECT schema_name FROM information_schema.schemata"
        ).fetchall()
        if (schema_name,) not in schemas:
            self.connection.sql(f"CREATE SCHEMA {schema_name}")

    def close(self) -> None:
        if self.connection is not None:
            self.connection.close()


class DuckDbParser(base.DatabaseParser):
    """Table parser for DuckDB schemas"""

    def parse_found_schema(self, schema: dict[str, str]) -> dict:
        """parses a information_schema.tables query response for valid columns"""
        parsed = {}

        for key, value in schema.items():
            if isinstance(value, str):
                # DuckDB provides a parser to go from string -> type objects
                value = duckdb.sqltypes.DuckDBPyType(value)

            # Collapse lists to the contained value
            if value.id == "list":
                value = value.children[0][1]  # [('child', CONTAINED_TYPE)]

            if value.id == "struct":
                result = self.parse_found_schema(dict(value.children))
            else:
                result = {}
            parsed[key.lower()] = result

        return parsed



"""Class for communicating with the Loinc API"""

import os
import pathlib

import requests
import rich

from cumulus_library import base_utils, errors

BASE_URL = "https://loinc.regenstrief.org/api/v1/"


class LoincApi:
    def __init__(self, *, user: str | None = None, password: str | None = None):
        """Creates a requests session for future calls

        :keyword user: the username of the loinc user
        :keyword password: the password of the loinc user

        You can request a Loinc account at https://loinc.org/join/.
        """
        if user is None:
            user = os.environ.get("LOINC_USER")
            if user is None:
                raise errors.ApiError("No LOINC user provided")
        if password is None:
            password = os.environ.get("LOINC_PASSWORD")
            if password is None:
                raise errors.ApiError("No LOINC password provided")

        self.session = requests.Session()
        self.session.auth = requests.auth.HTTPBasicAuth(user, password)
        self.cache_dir = base_utils.get_user_cache_dir() / "loinc"
        self.download_dir = self.cache_dir / "downloads"

    def get_all_download_versions(self) -> list:
        """returns all available versions available for download"""
        versions = []
        res = self.session.get(f"{BASE_URL}Loinc/All")
        if res.status_code == 401:
            raise errors.ApiError("Invalid LOINC credentials")
        for record in res.json():
            versions.append(record["version"])
        res.raise_for_status()
        return versions

    def get_download_info(self, version: str | None = None) -> tuple[str, str]:
        """gets the download info of the latest release, or the specified version
        :param version: a specific verson you'd like to download
        :returns: a tuple of the version (useful if not provided) and the download url
        """
        url = f"{BASE_URL}Loinc"
        if version is not None:
            url = f"{url}?version={version}"
        res = self.session.get(url)
        if res.status_code == 401:
            raise errors.ApiError("Invalid LOINC credentials")
        elif res.status_code == 404:
            raise errors.ApiError(f"Loinc version {version} not found")
        res.raise_for_status()
        res = res.json()
        return res["version"], res["downloadUrl"]

    def download_loinc_dataset(
        self,
        *,
        version: str | None = None,
        download_url: str | None = None,
        path: pathlib.Path | None = None,
        unzip: bool = True,
    ):
        """Downloads a dataset from the LOINC API
        :keyword version: the data version to download
        :keyword download_url: the url to download the zipfile from. Gets from API if not provided.
        :keyword path: the path on disk to write to (uses user cache dir if not provided)
        :keyword unzip: if True, extracts the archive after download (or a pre-existing download)
        """

        path = path or self.download_dir
        if download_url is None:
            version, download_url = self.get_download_info(version=version)
        path.mkdir(parents=True, exist_ok=True)

        if (path / f"{version}.zip").exists() or (path / version).exists():
            rich.print(f"Loinc version {version} already exists at {path}, skipping download")
        else:
            download_res = self.session.get(download_url, stream=True)
            with open(path / f"{version}.zip", "wb") as f:
                chunks_read = 0
                with base_utils.get_progress_bar() as progress:
                    task = progress.add_task(
                        f"Downloading {version}.zip",
                        total=(int(download_res.headers["Content-Length"]) / 1024),
                    )
                    for chunk in download_res.iter_content(chunk_size=1024):
                        f.write(chunk)
                        chunks_read += 1
                        progress.update(
                            task,
                            description=(f"Downloading {version}.zip: {chunks_read / 1000} MB"),
                            advance=1,
                        )
        if unzip and (path / f"{version}.zip").exists():
            (path / version).mkdir(parents=True, exist_ok=True)
            base_utils.unzip_file(path / f"{version}.zip", path / version)
            (path / f"{version}.zip").unlink()



"""Class for generating tables from a list of file uploads"""

import pathlib
import re

import pandas
import platformdirs

from cumulus_library import BaseTableBuilder, base_utils, errors, study_manifest
from cumulus_library.template_sql import base_templates

"""
A file upload config looks like this

type="file_upload"
[tables]
[table.name] # the table in db will be called 'name', snaked cased for sql compatibility
file = "relative/path/from/config/file.csv"
# The rest of these are optional
delimiter = "|" # comma by default, only applies to csv/tsv/bsv
col_types =["STRING","DATE","DOUBLE"] # default is all strings. Only relevant for Athena.
# See https://cwiki.apache.org/confluence/display/Hive/LanguageManual+Types for valid types
[table.other_name]
file = "other/path/file.csv"
"""


class FileUploadBuilder(BaseTableBuilder):
    display_text = "Uploading files..."

    @staticmethod
    def snake_case(x: str) -> str:
        """Converts either CAPS_CASE or CamelCase to snake_case"""
        return re.sub(r"([a-z])([A-Z])", r"\1_\2", x).lower()

    def __init__(self, *args, workflow_config: dict, toml_config_path: pathlib.Path, **kwargs):
        self.workflow_config = workflow_config
        self.toml_config_path = toml_config_path
        self.toml_config_dir = toml_config_path.parent
        super().__init__(*args, **kwargs)

    def prepare_queries(
        self,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest,
        *args,
        **kwargs,
    ):
        cache_dir = (
            pathlib.Path(platformdirs.user_cache_dir("cumulus-library", "smart-on-fhir"))
            / f"file_uploads/{manifest.get_study_prefix()}"
        )
        cache_dir.mkdir(parents=True, exist_ok=True)

        with base_utils.get_progress_bar() as progress:
            task = progress.add_task(
                "Uploading static files...",
                total=len(self.workflow_config["tables"]),
            )
            for table_name in self.workflow_config["tables"]:
                table = self.workflow_config["tables"][table_name]
                table_filename = pathlib.Path(table["file"]).name

                if table["file"].endswith(".xlsx"):
                    parquet_path = cache_dir / table_filename.replace(".xlsx", ".parquet")
                    df = pandas.read_excel(self.toml_config_dir / table["file"])

                elif (
                    table["file"].endswith(".csv")
                    or table["file"].endswith(".tsv")
                    or table["file"].endswith(".bsv")
                ):
                    parquet_path = cache_dir / f"{table_filename[:-4]}.parquet"
                    if "delimiter" not in table:
                        match table["file"].split(".")[-1]:
                            case "bsv":
                                table["delimiter"] = "|"
                            case "csv":
                                table["delimiter"] = ","
                            case "tsv":
                                table["delimiter"] = "\t"
                    df = pandas.read_csv(
                        self.toml_config_dir / table["file"], delimiter=table["delimiter"]
                    )

                elif table["file"].endswith(".parquet"):
                    parquet_path = cache_dir / table_filename
                    df = pandas.read_parquet(self.toml_config_dir / table["file"])

                else:
                    raise errors.FileUploadError(
                        f"{table['file']} is not a supported upload file type.\n"
                        "Supported file types: csv, bsv, tsv, xlsx, parquet"
                    )
                if "always_upload" not in table:
                    table["always_upload"] = True
                df = df.rename(self.snake_case, axis="columns")
                df.to_parquet(parquet_path)
                remote_path = config.db.upload_file(
                    file=parquet_path,
                    study=manifest.get_study_prefix(),
                    topic=parquet_path.stem,
                    force_upload=table["always_upload"] or config.force_upload,
                )
                if table.get("col_types") is None:
                    table["col_types"] = ["STRING" for x in df.columns]
                elif len(table["col_types"]) != len(df.columns):
                    raise errors.FileUploadError(
                        f"{table_name} has {len(df.columns)} columns, but the provided "
                        f"col_types has {len(table['col_types'])} entries."
                    )
                self.queries.append(
                    base_templates.get_ctas_from_parquet_query(
                        schema_name=config.schema,
                        table_name=f"{manifest.get_study_prefix()}__{table_name}",
                        local_location=parquet_path,
                        remote_location=remote_path,
                        table_cols=df.columns,
                        remote_table_cols_types=table["col_types"],
                    )
                )

                progress.advance(task)



import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "diagnosticreport": {
        "id": [],
        "status": [],
        "subject": sql_utils.REFERENCE,
        "encounter": sql_utils.REFERENCE,
        "effectiveDateTime": [],
        "effectivePeriod": ["start", "end"],
        "issued": [],
        "result": sql_utils.REFERENCE,
        "presentedForm": {
            "contentType": [],
            "data": [],
            "_data": {"extension": ["url", "valueCode"]},
        },
    }
}


class CoreDiagnosticReportBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating DiagnosticReport tables..."

    def prepare_queries(self, *args, config: cumulus_library.StudyConfig, **kwargs):
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(core_templates.get_core_template("diagnosticreport", validated_schema))



import dataclasses
import pathlib

from cumulus_library import base_utils, study_manifest


@dataclasses.dataclass(kw_only=True)
class ValuesetConfig:
    """Provides expected values for creating valuesets"""

    rules_file: str = None
    keyword_file: str = None
    table_prefix: str = None
    umls_stewards: dict[str, str] = None
    vsac_stewards: dict[str, str] = None


def get_valueset_cache_dir(
    path: pathlib.Path | None, manifest: study_manifest.StudyManifest | None
):
    # this is not actively used much right now, but if you want things to go
    # to the user cache dir, just don't provide a path
    if not path:  # pragma: no cover
        if manifest:
            subpath = f"{manifest.get_study_prefix()}/valueset_data"
        else:
            subpath = "valueset_data"  # pragma: no cover
        path = base_utils.get_user_cache_dir() / subpath
    # since a very common use case is 'download a file from vsac, make a flat file,
    # and then have the file uploader push it to S3', this is the default write location
    else:
        path = path / "valueset_data"
    path.mkdir(exist_ok=True, parents=True)
    return path



"""Module for generating core medicationrequest table"""

import cumulus_library
from cumulus_library.template_sql import sql_utils


class MedicationRequestBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating MedicationRequest table..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ) -> None:
        """Constructs queries related to medication requests

        :param config: A study config object
        """
        code_sources = [
            sql_utils.CodeableConceptConfig(
                source_table="medication",
                column_hierarchy=[("code", dict)],
                target_table="core__medication_dn_code",
            ),
            sql_utils.CodeableConceptConfig(
                source_table="medicationrequest",
                column_hierarchy=[("medicationCodeableConcept", dict)],
                target_table="core__medicationrequest_dn_inline_code",
            ),
            sql_utils.CodeableConceptConfig(
                source_table="medicationrequest",
                column_hierarchy=[("contained", list), ("code", dict)],
                target_table="core__medicationrequest_dn_contained_code",
                expected={
                    "code": sql_utils.CODEABLE_CONCEPT,
                    "id": {},
                    "resourceType": {},
                },
                extra_fields=[
                    ("id", "contained_id"),
                    ("resourceType", "resource_type"),
                ],
            ),
            sql_utils.CodeableConceptConfig(
                source_table="medicationrequest",
                column_hierarchy=[("category", list)],
                target_table="core__medicationrequest_dn_category",
            ),
        ]
        self.queries += sql_utils.denormalize_complex_objects(
            config.db, code_sources, "MedicationRequest"
        )



"""Module for generating encounter codeableConcept table"""

import cumulus_library
from cumulus_library.studies.discovery import code_definitions


class CleanupTmpBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Selecting unique code systems..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        """Constructs queries related to condition codeableConcept

        :param config: A study config object
        """

        code_sources = []
        for code_definition in code_definitions.code_list:
            code_source = {}
            for key in code_definition.keys():
                code_source[key] = code_definition[key]
            code_sources.append(code_source)
        query = cumulus_library.get_template(
            "show_tables", schema_name=config.schema, prefix="discovery__tmp"
        )
        tmp_tables = config.db.connection.execute(query).fetchall()
        for table in tmp_tables:
            query = cumulus_library.get_template(
                "drop_view_table", view_or_table="table", view_or_table_name=table[0]
            )
            self.queries.append(query)



from dataclasses import dataclass
from pathlib import Path

from cumulus_library import errors
from cumulus_library.template_sql import base_templates


@dataclass
class CountColumn:
    name: str
    db_type: str
    alias: str


@dataclass(kw_only=True)
class FilterColumn:
    name: str
    values: list[str]
    include_nulls: bool

    # alias is only included for jinja compatibility reasons with CountColumn
    @property
    def alias(self) -> str | None:
        return None  # pragma: no cover


@dataclass(kw_only=True)
class CountAnnotation:
    """Defines a table to use as a source for annotating a count dataset.

    :keyword field: the field in the powerset source table to use for the join.
        Note: in most cases, this field should not be included in your powerset.
    :keyword join_table: the name of the table to join from. If the table is
        in a different schema, use '"schema"."table' as the format
    :keyword join_field: the column in the table to join on
    :keyword columns: a list of tuples like ('column_name', 'alias' or None) to
        define columns to add to your dataset.
    :keyword alt_target: if present, a column to use from the annotation table
        instead of the one defined in `field` as the primary column in the
        annotated power set. This is intended to be the column representing
        the field of highest granularity, but functionally treats the combination
        of all fields in columns as a unique key.

    """

    field: str
    join_table: str
    join_field: str
    columns: list[str] | list[tuple[str, str, str | None]] | list[CountColumn]
    alt_target: str | None = None


def _cast_table_col(col):
    if isinstance(col, str):
        return CountColumn(name=col, db_type="VARCHAR", alias=None)
    elif isinstance(col, CountColumn):
        return col
    else:
        return CountColumn(name=col[0], db_type=col[1], alias=col[2])


def _cast_filter_col(filter_col):
    if isinstance(filter_col, tuple) or isinstance(filter_col, list):
        return FilterColumn(name=filter_col[0], values=filter_col[1], include_nulls=filter_col[2])
    elif isinstance(filter_col, FilterColumn):
        return filter_col


def get_count_query(
    table_name: str,
    source_table: str,
    table_cols: str | list[list[str] | CountColumn],
    *args,
    min_subject: int = 10,
    where_clauses: list[str] | None = None,
    primary_id: str | None = None,
    secondary_id: str | None = None,
    alt_secondary_join_id: str | None = None,
    secondary_table: str | None = None,
    secondary_cols: list[str] = [],
    patient_link: str | None = None,  # deprecated legacy arg, v6.0.0
    annotation: CountAnnotation | None = None,
    filter_status: bool | None = False,
    filter_cols: list[tuple[str, list[str], bool]] | list[FilterColumn] = [],
    **kwargs,
) -> str:
    """Generates count tables for generating study outputs"""

    if primary_id is None:
        if patient_link:  # pragma: no cover
            primary_id = patient_link
        else:
            primary_id = "subject_ref"
    path = Path(__file__).parent

    # we are going to paper over a couple of dataclass vs dicts and lists interactions here to allow
    # for differing levels of user complexity - but the end goal is to get data out in one form for
    # template management reasons
    table_col_classed = []
    for item in table_cols:
        table_col_classed.append(_cast_table_col(item))

    table_cols = table_col_classed
    if annotation:
        annotation_col_classed = []
        for item in annotation.columns:
            annotation_col_classed.append(_cast_table_col(item))
        annotation.columns = annotation_col_classed
    if filter_status and len(filter_cols) == 0:
        raise errors.CountsBuilderError(  # pragma: no cover
            "When filtering in a CountsBuilder, both 'filter_status' and "
            "'filter_cols' must be supplied."
        )
    filter_cols_classed = []
    if filter_cols:
        for filter_col in filter_cols:
            filter_cols_classed.append(_cast_filter_col(filter_col))
    filter_cols = filter_cols_classed
    query = base_templates.get_template(
        "count",
        path,
        table_name=table_name,
        primary_table=source_table,
        table_cols=table_cols,
        min_subject=min_subject,
        where_clauses=where_clauses,
        primary_id=primary_id,
        secondary_table=secondary_table,
        secondary_id=secondary_id,
        alt_secondary_join_id=alt_secondary_join_id,
        secondary_cols=secondary_cols,
        annotation=annotation,
        filter_status=filter_status,
        filter_cols=filter_cols,
    )
    # workaround for conflicting sqlfluff enforcement
    return query.replace("-- noqa: disable=LT02\n", "")



import pathlib
import sys
from collections.abc import Iterable
from concurrent import futures

import cumulus_fhir_support
import pyarrow
import pyarrow.dataset
import pyarrow.json
import rich

from cumulus_library import base_utils, db_config, errors
from cumulus_library.databases import athena, base, duckdb


def _list_files_for_resource(path: pathlib.Path, resource: str) -> list[str]:
    return list(cumulus_fhir_support.list_multiline_json_in_dir(path, resource, recursive=True))


def _rows_from_files(files: list[str]) -> Iterable[dict]:
    for file in files:
        yield from cumulus_fhir_support.read_multiline_json(file)


def _json_format():
    """Returns a Dataset format object suitable for reading FHIR NDJSON."""
    # FHIR can have very long JSON lines (think DocRefs with inlined notes). And unfortunately,
    # the way PyArrow's JSON parser works is that each entire line must fit inside a single read
    # block, or an exception is raised. I've seen 13MB JSON lines in the wild, at least.
    # So let's be generous here and allow 50MB lines. And cross our fingers that it's enough.
    read_options = pyarrow.json.ReadOptions(block_size=50 * 2**20)
    return pyarrow.dataset.JsonFileFormat(read_options=read_options)


def _load_custom_etl_table(files: list[str]) -> pyarrow.dataset.Dataset | None:
    """Loads a non-FHIR ETL table from disk (tables like etl__completion)."""

    # This is a custom Cumulus ETL table, with no easy way to get a schema definition.
    # We **could** put a hard-coded shared schema in cumulus-fhir-support, but since these
    # tables are so simple, we mostly don't yet need to have that level of pre-coordination.
    #
    # Instead, we let PyArrow infer the types, with one hiccup: its JSON parser will interpret
    # datetime strings as a TIMESTAMP type, which isn't insane, but not what we want - we want to
    # keep those as strings. So we let PyArrow infer from Python objects, where it works like we
    # want.
    #
    # Thus, we sip on the data a bit to infer a schema from the first row as a Python object.
    first_row = next(iter(_rows_from_files(files)), None)
    if not first_row:
        # This path shouldn't happen, since list_multiline_json_in_dir() won't give us back
        # empty files. But just in case.
        return None  # pragma: no cover
    schema = pyarrow.Table.from_pylist([first_row]).schema

    # Now let PyArrow load the rest of the data from disk on demand, rather than loading it all
    # into memory, but using the detected schema rather than inferring it with timestamps.
    return pyarrow.dataset.dataset(files, schema=schema, format=_json_format())


def get_ndjson_files(path: str) -> dict[dict[str]]:
    """Gets a list of expected possible FHIR ndjson from the target path"""

    all_tables = {}
    resources = {}
    metadata = {}
    # Manually specify the list of resources because we want to create each table
    # even if the folder does not exist.
    resource_names = [
        "AllergyIntolerance",
        "Condition",
        "Device",
        "DiagnosticReport",
        "DocumentReference",
        "Encounter",
        "Immunization",
        "Location",
        "Medication",
        "MedicationRequest",
        "Observation",
        "Organization",
        "Patient",
        "Practitioner",
        "PractitionerRole",
        "Procedure",
        "ServiceRequest",
    ]
    for resource in resource_names:
        files = _list_files_for_resource(pathlib.Path(path), resource)
        resources[resource] = files
    all_tables["resources"] = resources

    metadata_tables = [
        "etl__completion",
        "etl__completion_encounters",
    ]
    for metadata_table in metadata_tables:
        files = list(
            cumulus_fhir_support.list_multiline_json_in_dir(pathlib.Path(path) / metadata_table)
        )
        metadata[metadata_table] = files
    all_tables["metadata"] = metadata
    return all_tables


def read_ndjson_dir(
    path: str | None, fileset: dict[dict[str]] | None = None
) -> dict[str, pyarrow.dataset.Dataset]:
    """Loads a directory tree of raw ndjson into schema-ful tables.

    One of either path or fileset must be provided.

    :param path: a directory path to ndjson
    :param fileset: A dictionary of ndjson file locations (usually the output of
      get_ndjson_files)
    :returns: dictionary of table names (like 'documentreference') to table
      data (with schema)
    """
    all_tables = {}
    if not fileset and not path:
        raise errors.CumulusLibraryError(
            "databases.utils.read_ndjson_dir() requires either a path or fileset argument."
        )
    # You may already have a fileset from using get_ndjson_files() to check if
    # your FHIR data has changed, if your DB is using that to cache.
    # If not, we'll just get that list directly.
    if not fileset:
        fileset = get_ndjson_files(path)

    for resource, files in fileset["resources"].items():
        table_name = resource.lower()
        # Make a pyarrow table with full schema from the data
        schema = cumulus_fhir_support.pyarrow_schema_from_rows(resource, _rows_from_files(files))
        # Use a PyArrow Dataset (vs a Table) to avoid loading all the files in memory.
        all_tables[table_name] = pyarrow.dataset.dataset(
            files, schema=schema, format=_json_format()
        )

    # And now some special support for a few ETL tables.
    for metadata_table, meta_path in fileset["metadata"].items():
        if dataset := _load_custom_etl_table(meta_path):
            all_tables[metadata_table] = dataset
    return all_tables


def _handle_load_ndjson_dir(args: dict[str, str], backend: base.DatabaseBackend) -> None:
    load_ndjson_dir = args.get("load_ndjson_dir")
    if not load_ndjson_dir:
        return  # nothing to do

    if db_config.db_type != "duckdb":
        sys.exit("Loading an NDJSON dir is only supported with --db-type=duckdb.")

    if backend.connection is None:
        return  # connect() was never run, we don't have a live DB connection

    with base_utils.get_progress_bar() as progress:
        progress.add_task("Detecting JSON schemas...", total=None)
        tables = get_ndjson_files(load_ndjson_dir)
        backend.insert_tables(tables)


def create_db_backend(
    args: dict[str, str], pyarrow_cache_path: str | None = None
) -> tuple[base.DatabaseBackend, str]:
    """Retrieves a database backend and target schema from CLI args"""
    db_config.db_type = args["db_type"]

    if db_config.db_type == "duckdb":
        # TODO: reevaluate as DuckDB's local schema support evolves.
        # https://duckdb.org/docs/sql/statements/set.html#syntax
        if not (args.get("schema_name") is None or args["schema_name"] == "main"):
            rich.print(  # pragma: no cover
                "Warning - local schema names are not yet supported by duckDB's "
                "python library - using 'main' instead"
            )
        schema_name = "main"
        backend = duckdb.DuckDatabaseBackend(
            args["database"],
            max_concurrent=args.get("max_concurrent"),
            pyarrow_cache_path=pyarrow_cache_path,
        )
    elif db_config.db_type == "athena":
        if (
            args.get("schema_name") is not None
            and args.get("database") is not None
            and args.get("schema_name") != args.get("database")
        ):
            sys.exit(
                f"Two separate values, database: '{args['database']}' and "
                f"schema_name: '{args['schema_name']}' were supplied for "
                "an Athena database connection.\n"
                "Only one of these needs to be supplied. Both can be passed "
                "as arguments, but they must be equal"
            )
        schema_name = args["schema_name"] or args["database"]
        backend = athena.AthenaDatabaseBackend(
            args["region"],
            args["work_group"],
            args["profile"],
            schema_name,
            args.get("max_concurrent"),
        )
    else:
        raise errors.CumulusLibraryError(f"'{db_config.db_type}' is not a supported database.")
    if "prepare" not in args.keys():
        backend.connect()
    elif not args["prepare"]:
        backend.connect()
    _handle_load_ndjson_dir(args, backend)
    return (backend, schema_name)


def handle_concurrent_errors(
    resolved_futures: list[tuple[str, futures.Future]], db_type: str
) -> None:
    """Exits with formatted errors if a query fails during parallel execution"""
    failures = []
    for f in resolved_futures:
        try:
            res = f[1].result()
            # pyathena has a helper class for resolved futures,
            # so we'll check for it's specific failure indicator
            if db_type == "athena":
                if res.state == "FAILED":
                    failures.append((f[0], res.error_message))
        except Exception:
            failures.append((f[0], str(f[1].exception())))
    if failures:
        exit_msg = ""
        for f in failures:
            exit_msg = f"{exit_msg}\n-----\n{f[0]}\n\nHad the following error:\n\n"
            exit_msg += f"{f[1]}\n"
        sys.exit(f"One or more queries failed to execute:\n{exit_msg}")



"""Collection of small commonly used utility functions"""

import dataclasses
import datetime
import pathlib
import shutil
import zipfile
from contextlib import contextmanager

import platformdirs
import rich
from rich import progress

from cumulus_library import databases, study_manifest


@dataclasses.dataclass
class StudyConfig:
    """Class for containing study-centric parameters

    The intent of this class is that if you want something passed through to the
    prepare_queries section of a study, this is the place it should go. If you're
    doing something above that level, consider explicit arguments instead. This should
    be an interface aimed at a study author.

    :param db: a databaseBackend object for a specific target database
    :param schema: the database schema specified at the command line
    :keyword drop_table: when creating tables, if one already exists in the db,
        drop it
    :keyword force_upload: If the study downloads data from an external resource,
        force it to skip any cached data when running
    :keyword verbose: if True, print raw queries to the cli instead of progress bars
    :keyword stats_build: If the study runs a stats builder, force regeneration of
        any sampling or other stochastic techniques
    :keyword umls_key: An API for the UMLS service, used for downloading vocabularies
    :keyword umls: A UMLS API key
    :keyword options: a dictionary for any study-specific CLI arguments
    """

    db: databases.DatabaseBackend
    schema: str
    drop_table: bool = False
    force_upload: bool = False
    verbose: bool = False
    stats_build: bool = False
    stats_clean: bool = False
    umls_key: str | None = None
    loinc_user: str | None = None
    loinc_password: str | None = None
    options: dict | None = None


def get_schema(config: StudyConfig, manifest: study_manifest.StudyManifest):
    if dedicated := manifest.get_dedicated_schema():
        config.db.create_schema(dedicated)
        return dedicated
    return config.schema


def load_text(path: str) -> str:
    with open(path, encoding="UTF-8") as fp:
        return fp.read()


def parse_sql(sql_text: str) -> list[str]:
    commands = []

    for statement in sql_text.split(";"):
        parsed = []
        for line in statement.splitlines():
            if not line.strip().startswith("--"):
                parsed.append(line)
        commands.append("\n".join(parsed))
    return filter_strip(commands)


def filter_strip(commands) -> list[str]:
    return list(filter(None, [c.strip() for c in commands]))


@contextmanager
def query_console_output(
    verbose: bool, query: str, progress_bar: progress.Progress, task: progress.Task
):
    """Convenience context manager for handling console output"""
    if verbose:
        rich.print()
        rich.print(query)
    yield
    if not verbose:
        progress_bar.advance(task)


def get_progress_bar(
    bar_type=progress.TaskProgressColumn(), metric=progress.TimeElapsedColumn(), **kwargs
) -> progress.Progress:
    # The default columns don't change to elapsed time when finished.
    return progress.Progress(
        progress.TextColumn("[progress.description]{task.description}"),
        progress.BarColumn(),
        bar_type,
        metric,
        **kwargs,
    )


def get_utc_datetime() -> datetime.datetime:
    return datetime.datetime.now(datetime.UTC).replace(microsecond=0)


def get_tablename_safe_iso_timestamp() -> str:
    """formats a timestamp to remove sql unallowed characters in table names"""
    iso_timestamp = get_utc_datetime().isoformat()
    safe_timestamp = iso_timestamp.replace(":", "_").replace("-", "_").replace("+", "_")
    return safe_timestamp


def zip_dir(read_path, write_path, archive_name, archive_csvs=False):
    """Moves a directory to an archive"""
    file_list = [file for file in read_path.glob("**/*") if file.is_file()]
    timestamp = get_utc_datetime().isoformat().replace("+00:00", "Z")
    if archive_csvs:
        # archives including csvs are meant to be permanent and are kept outside the study data dirs
        archive_path = f"{write_path}/{archive_name}__{timestamp}.zip"
    else:
        # otherwise archives are for upload and are ephemeral, and will be replaced each export
        archive_path = f"{write_path}/{archive_name}/{archive_name}.zip"
    with zipfile.ZipFile(
        archive_path,
        "w",
        zipfile.ZIP_DEFLATED,
    ) as f:
        for file in file_list:
            if (not file.suffix == ".csv" and archive_csvs is False) or archive_csvs is True:
                f.write(file, file.relative_to(read_path))
                file.unlink()
        if archive_csvs:
            shutil.rmtree(read_path)


def update_query_if_schema_specified(query: str, manifest: study_manifest.StudyManifest):
    if manifest and manifest.get_dedicated_schema():
        # External queries in athena require a schema to be specified already, so
        # rather than splitting and ending up with a table name like
        # `schema`.`schema.table`, we just remove the `schema__` chunk from the
        # table name.
        # TODO: Move this to be a database defined function.
        if "CREATE EXTERNAL" in query:
            query = query.replace(
                f"{manifest.get_study_prefix()}__",
                "",
            )
        else:
            query = query.replace(
                f"{manifest.get_study_prefix()}__",
                f"{manifest.get_dedicated_schema()}.",
            )
    return query


def unzip_file(file_path: pathlib.Path, write_path: pathlib.Path):
    """Expands a zip archive"""
    with zipfile.ZipFile(file_path, mode="r") as z:
        for file in z.namelist():
            z.extract(file, write_path)


def get_user_cache_dir() -> pathlib.Path:
    return pathlib.Path(platformdirs.user_cache_dir("cumulus-library", "smart-on-fhir"))


def get_user_documents_dir() -> pathlib.Path:
    return pathlib.Path(platformdirs.user_documents_dir())



import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "encounter": {
        "id": [],
        "status": [],
        "period": [
            "start",
            "end",
        ],
        "class": sql_utils.CODING,
        "subject": sql_utils.REFERENCE,
        "participant": {
            "individual": sql_utils.REFERENCE,
        },
        "serviceProvider": sql_utils.REFERENCE,
    },
    "etl__completion": {
        "group_name": [],
    },
    "etl__completion_encounters": {
        "group_name": [],
    },
}


class CoreEncounterBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Encounter tables..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries += [
            core_templates.get_core_template("encounter", validated_schema),
            core_templates.get_core_template("incomplete_encounter", validated_schema),
        ]



from .athena import AthenaDatabaseBackend, AthenaParser
from .base import DatabaseBackend, DatabaseCursor, DatabaseParser
from .duckdb import DuckDatabaseBackend, DuckDbParser
from .utils import create_db_backend, get_ndjson_files, read_ndjson_dir



import typing

import pandas
import pytablewriter

from cumulus_library import base_utils, databases, study_manifest
from cumulus_library.actions import builder
from cumulus_library.template_sql import base_templates


def run_generate_sql(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    table_builder: str | None = None,
    db_parser: databases.DatabaseParser = None,
) -> None:
    """Generates reference SQL from table_builders listed in the manifest

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword table_builder: a specific table_builder to target
    :keyword db_parser: a DB parser
    """
    all_generators = manifest.get_all_generators()
    doc_str = (
        "-- This sql was autogenerated as a reference example using the library\n"
        "-- CLI. Its format is tied to the specific database it was run against,\n"
        "-- and it may not be correct for all databases. Use the CLI's build \n"
        "-- option to derive the best SQL for your dataset."
    )
    for run_stage, file in enumerate(all_generators):
        if table_builder and file.find(table_builder) == -1:
            continue  # pragma: no cover
        builder._run_builder(
            config=config,
            manifest=manifest,
            filename=file,
            db_parser=db_parser,
            write_reference_sql=True,
            doc_str=doc_str,
            data_path=None,
            run_stage=run_stage,
        )


def run_generate_markdown(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    db_parser: databases.DatabaseParser = None,
) -> None:
    """Generates reference SQL from builders listed in the manifest

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    """

    cursor = config.db.cursor()
    query = base_templates.get_show_tables(
        schema_name=config.schema, prefix=f"{manifest.get_study_prefix()}__"
    )
    tables = [x[0] for x in cursor.execute(query).fetchall()]
    query = base_templates.get_column_datatype_query(
        schema_name=config.schema, table_names=tables, include_table_names=True
    )
    study_df = pandas.DataFrame(
        cursor.execute(query).fetchall(), columns=["Column", "Type", "Table"]
    )
    with open(
        manifest._study_path / f"{manifest.get_study_prefix()}_generated.md",
        "w",
    ) as f:
        table_list = sorted(study_df["Table"].unique())
        count_tables = [t for t in table_list if "__count_" in t]
        base_tables = [t for t in table_list if "__count_" not in t]
        if len(count_tables) > 0:
            f.write(f"## {manifest.get_study_prefix()} count tables\n\n")
            for table in count_tables:
                _write_md_table(table, study_df, f)
        if len(base_tables) > 0:
            f.write(f"## {manifest.get_study_prefix()} base tables\n\n")
            for table in base_tables:
                _write_md_table(table, study_df, f)


def _write_md_table(name: str, df: pandas.DataFrame, file: typing.IO):
    table_df = df[df["Table"] == name].drop("Table", axis=1)
    table_df = table_df.assign(Description="")
    writer = pytablewriter.MarkdownTableWriter(dataframe=table_df)
    writer.table_name = f"{name}\n"
    writer.set_indent_level(2)
    writer.stream = file
    writer.write_table()
    file.write("\n\n")



import pathlib
import tempfile
import zipfile

import pandas

from cumulus_library import base_utils, errors
from cumulus_library.actions import cleaner
from cumulus_library.template_sql import base_templates


def _create_table_from_parquet(archive, file, study_name, config):
    try:
        parquet_path = pathlib.Path(archive.extract(file), path=tempfile.TemporaryFile())
        # While convenient to access, this exposes us to panda's type system,
        # which is messy - this could be optionally be replaced by pyarrow if it
        # becomes problematic.
        table_types = pandas.read_parquet(parquet_path).dtypes
        remote_types = config.db.col_parquet_types_from_pandas(table_types.values)
        s3_path = config.db.upload_file(
            file=parquet_path,
            study=study_name,
            topic=parquet_path.stem,
            force_upload=True,
            remote_filename=parquet_path.name,
        )
        query = base_templates.get_ctas_from_parquet_query(
            schema_name=config.schema,
            table_name=parquet_path.stem.replace(".", "_"),
            local_location=f"{parquet_path.parent}/*.parquet",
            remote_location=s3_path,
            table_cols=list(table_types.index),
            remote_table_cols_types=remote_types,
        )
        config.db.cursor().execute(query)
    finally:
        parquet_path.unlink()


def import_archive(config: base_utils.StudyConfig, *, archive_path: pathlib.Path):
    """Creates a study in the database from a previous export

    :param config: a StudyConfig object
    :keyword archive_path: the location of the archive to import data from

    """

    # Ensure we've got something that looks like a valid database export
    if not archive_path.exists():
        raise errors.StudyImportError(f"File {archive_path} not found.")
    try:
        archive = zipfile.ZipFile(archive_path)
        files = archive.namelist()
        files = [file for file in files if file.endswith(".parquet")]
    except zipfile.BadZipFile as e:
        raise errors.StudyImportError(f"File {archive_path} is not a valid archive.") from e
    if not any("__" in file for file in files):
        raise errors.StudyImportError(f"File {archive_path} contains non-study parquet files.")
    study_name = files[0].split("__")[0]
    for file in files[1:]:
        if file.split("__")[0] != study_name:
            raise errors.StudyImportError(
                f"File {archive_path} contains data from more than one study."
            )

    # Clean and rebuild from the provided archive
    cleaner.clean_study(config=config, manifest=None, prefix=study_name)
    with base_utils.get_progress_bar(disable=config.verbose) as progress:
        task = progress.add_task(
            f"Recreating {study_name} from archive...",
            total=len(files),
            visible=not config.verbose,
        )
        for file in files:
            _create_table_from_parquet(archive, file, study_name, config)
            progress.advance(task)



"""Builder for generating subsets of RxNorm data from a given valueset"""

import pathlib

from cumulus_library import BaseTableBuilder, base_utils, study_manifest
from cumulus_library.builders.valueset import valueset_utils
from cumulus_library.template_sql import base_templates


class AdditionalRulesBuilder(BaseTableBuilder):
    display_text = "Generating rulesets..."
    base_path = pathlib.Path(__file__).resolve().parent

    def prepare_queries(
        self,
        *args,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest,
        valueset_config: valueset_utils.ValuesetConfig,
        **kwargs,
    ):
        self.parallel_allowed = False
        study_prefix = manifest.get_prefix_with_seperator()
        table_prefix = ""
        if valueset_config.table_prefix:
            table_prefix = f"{valueset_config.table_prefix}_"
        self.queries.append(
            base_templates.get_base_template(
                "create_search_rules_descriptions",
                self.base_path / "template_sql",
                study_prefix=study_prefix,
                table_prefix=table_prefix,
            )
        )
        self.queries.append(
            # Get all matching rules from the keywords table that match rels
            base_templates.get_create_table_from_tables(
                table_name=f"{study_prefix}{table_prefix}potential_rules",
                # From a domain logic perspective, the _rela table is
                # the leftmost table and we're annotating with the
                # data from rxnconso. Since rxnconso is much, much
                # larger, we're moving it to the left in the actual
                # constructed join for athena performance reasons
                tables=[
                    f"{study_prefix}{table_prefix}all_rxnconso_keywords",
                    f"{study_prefix}{table_prefix}rela",
                ],
                table_aliases=["r", "s"],
                columns=[
                    "s.rxcui",
                    "r.rxcui",
                    "s.tty",
                    "r.tty",
                    "s.rui",
                    "s.rel",
                    "s.rela",
                    "s.str",
                    "r.str",
                    "r.keyword",
                ],
                column_aliases={
                    "s.rxcui": "rxcui1",
                    "s.tty": "tty1",
                    "s.str": "str1",
                    "r.rxcui": "rxcui2",
                    "r.tty": "tty2",
                    "r.str": "str2",
                },
                join_clauses=[
                    "s.rxcui2 = r.rxcui",
                    (
                        "s.rxcui2 NOT IN (SELECT DISTINCT RXCUI FROM "  # noqa: S608
                        f"{study_prefix}{table_prefix}rxnconso_keywords)"
                    ),
                ],
            )
        )
        # with the uploaded search rules, get all matching rels that line up with one
        # of the search rule types
        self.queries.append(
            base_templates.get_create_table_from_tables(
                table_name=f"{study_prefix}{table_prefix}included_rels",
                tables=[
                    f"{study_prefix}{table_prefix}potential_rules",
                    f"{study_prefix}{table_prefix}search_rules",
                ],
                table_aliases=["r", "e"],
                columns=[
                    "r.rxcui1",
                    "r.rxcui2",
                    "r.tty1",
                    "r.tty2",
                    "r.rui",
                    "r.rel",
                    "r.rela",
                    "r.str1",
                    "r.str2",
                    "r.keyword",
                ],
                join_clauses=[
                    "r.REL NOT IN ('RB', 'PAR')",
                    "e.include = TRUE",
                    "r.TTY1 = e.TTY1",
                    "r.TTY2 = e.TTY2",
                    "r.RELA = e.RELA",
                ],
            )
        )
        # And then do the same thing, but look for keyword presence
        self.queries.append(
            base_templates.get_base_template(
                "create_included_keywords",
                self.base_path / "template_sql",
                study_prefix=study_prefix,
                table_prefix=table_prefix,
            )
        )
        # Then combine all keyword matches with all rel matches for the
        # final ruleset table
        self.queries.append(
            base_templates.get_create_table_from_union(
                table_name=f"{study_prefix}{table_prefix}combined_ruleset",
                tables=[
                    f"{study_prefix}{table_prefix}included_keywords",
                    f"{study_prefix}{table_prefix}included_rels",
                ],
                columns=[
                    "rxcui1",
                    "rxcui2",
                    "tty1",
                    "tty2",
                    "rui",
                    "rel",
                    "rela",
                    "str1",
                    "str2",
                    "keyword",
                ],
            )
        )



"""Module for generating age ranges table from NLP"""

import pathlib

import cumulus_library
from cumulus_library import base_utils, databases
from cumulus_library.template_sql import sql_utils


class CalculateRangesBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Calculating age ranges..."

    @staticmethod
    def _is_table_valid(database: databases.DatabaseBackend, table_name: str) -> bool:
        return sql_utils.is_field_present(
            database=database,
            source_table=table_name,
            source_col="note_ref",
            expected=[],
        )

    def _get_valid_nlp_tables(self, database: databases.DatabaseBackend) -> set[str]:
        source_tables = [
            "example_nlp__nlp_gpt4",
            "example_nlp__nlp_gpt4o",
            "example_nlp__nlp_gpt5",
            "example_nlp__nlp_gpt_oss_120b",
            "example_nlp__nlp_llama4_scout",
        ]
        valid_tables = set()
        with base_utils.get_progress_bar() as progress:
            task = progress.add_task(
                "Discovering available NLP tables...",
                total=len(source_tables),
            )
            for source_table in source_tables:
                if self._is_table_valid(database, source_table):
                    valid_tables.add(source_table)
                progress.advance(task)
        return valid_tables

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        valid_tables = self._get_valid_nlp_tables(config.db)
        query = cumulus_library.get_template(
            "calculate_ranges",
            pathlib.Path(__file__).parent,
            table_names=valid_tables,
        )
        self.queries.append(query)



"""utility functions related to populating jinja templates

Currently, these deal with various edge cases around complex nested objects in sql
and validating them so that the actual generated queries can be constructed more
simply. This includes, but is not limited, to:
    - Data missing entirely
    - Data present, but 'nullish' - with some structure but no value
    - Data with deep missing elements
    - Data which may or may not be in an array depending on context
"""

from dataclasses import dataclass, field

from cumulus_library import base_utils, databases, enums
from cumulus_library.template_sql import base_templates

# *** Some convenience constants for providing to validate_schema() ***
# * Only include necessary fields that we actually use.
# * If any sub-field is not present, we may ignore the parent-field.
# * Though, cumulus-fhir-support does guarantee a full Coding or Concept.
CODING = ["code", "display", "system", "userSelected", "version"]
# If you think you need CODEABLE_CONCEPT, you probably need a de-normalization table
# instead, since CodeableConcepts can contain multiple entries.
CODEABLE_CONCEPT = {"coding": CODING, "text": {}}
REFERENCE = ["reference"]


@dataclass(kw_only=True)
class BaseTable:
    """Base class for system table metadata

    :keyword name: the name of the table
    :keyword columns: a list of columns names
    :keyword column_types: a list of sql types for the columns
    :keyword type_casts: a dict of column/type values that require explicit casting
    """

    name: str
    columns: list
    column_types: list
    type_casts: dict


@dataclass(kw_only=True)
class TransactionsTable(BaseTable):
    name: str = enums.ProtectedTables.TRANSACTIONS.value
    columns: list = field(
        default_factory=lambda: [
            "study_name",
            "library_version",
            "status",
            "event_time",
            "message",
        ]
    )
    column_types: list = field(
        default_factory=lambda: [
            "varchar",
            "varchar",
            "varchar",
            "timestamp",
            "varchar",
        ]
    )
    type_casts: dict = field(default_factory=lambda: {"event_time": "timestamp"})


@dataclass(kw_only=True)
class StatisticsTable(BaseTable):
    name: str = enums.ProtectedTables.STATISTICS.value
    columns: list = field(
        default_factory=lambda: [
            "study_name",
            "library_version",
            "table_type",
            "table_name",
            "view_name",
            "created_on",
        ]
    )
    column_types: list = field(
        default_factory=lambda: [
            "varchar",
            "varchar",
            "varchar",
            "varchar",
            "varchar",
            "timestamp",
        ]
    )
    type_casts: dict = field(default_factory=lambda: {"created_on": "timestamp"})


@dataclass(kw_only=True)
class BaseFHIRResourceConfig:
    """Base class for handling table detection/denormalization"""

    source_table: str = None
    source_id: str = "id"
    target_table: str = None
    has_data: bool = False


@dataclass(kw_only=True)
class CodeableConceptConfig(BaseFHIRResourceConfig):
    """Holds parameters for generating codableconcept tables.

    :keyword column_name: the column containing the codeableConcept you want to extract.
    :keyword is_array: whether the codeableConcept is 0...1 or 0..* in the FHIR spec
    :keyword source_table: the table to extract extensions from
    :keyword target_table: the name of the table to create
    :keyword source_id: the id field to use in the new table (default: 'id')
    :keyword filter_priority: If true, will use code systems to select a single code,
      in preference order, for use as a display value.
    :keyword code_systems: a list of strings matching the start of the systems field,
      in preference order, for selecting data for filtering. This should not be set
      if filter_priority is false.
    :keyword expected: a schema fragment for what the shape of this field should be.
      If any bit of this schema fragment is not present, there will be no results.
    :keyword extra_fields: extra fields to include, as siblings of the concept.
      It's a list of tuples (field, alias).
    """

    column_hierarchy: list[tuple]
    filter_priority: bool = False
    code_systems: list = None
    expected: list | dict = field(default_factory=lambda: CODEABLE_CONCEPT)
    extra_fields: list[tuple] = None  # candidate for moving into base config


@dataclass(kw_only=True)
class CodingConfig(BaseFHIRResourceConfig):
    column_hierarchy: list[tuple]
    filter_priority: bool = False
    code_systems: list = None
    expected: list | dict = field(default_factory=lambda: CODING)


@dataclass(kw_only=True)
class ExtensionConfig(BaseFHIRResourceConfig):
    """convenience class for holding parameters for generating extension tables.

    :keyword source_table: the table to extract extensions from
    :keyword source_id: the id column to treat as a foreign key
    :keyword target_table: the name of the table to create
    :keyword target_col_prefix: the string to prepend code/display column names with
    :keyword fhir_extension: the URL of the FHIR resource to select
    :keyword ext_systems: a list of codes, in preference order, to use to select data
    :keyword is_array: a boolean indicating if the targeted field is an array type
    """

    target_col_prefix: str
    fhir_extension: str
    ext_systems: list[str]
    is_array: bool = False


def _check_data_in_fields(
    database: databases.DatabaseBackend,
    code_sources: list[CodeableConceptConfig],
    resource: str | None = None,
) -> dict:
    """checks if CodeableConcept fields actually have data available

    CodeableConcept fields are mostly optional in the FHIR spec, and may be arrays
    or single objects. Additionally, the null representation can be inconsistent,
    depending on how the data is provided from an EHR and how the ETL manages
    schema inference (wide, but not deep). We :could: try to find the data and
    just catch an error, but that would mask configuration errors/unexpected
    data patterns. So, instead, we are doing the following fussy operation:

    For each column we want to check for data:
    - Check to see if there is any data in a codeableConcept field
    - Check to see if the codeableConcept field contains a coding element
    - Check if that coding element contains non-null data

    The way we do this is slightly different depending on if the field is an
    array or not (generally requiring one extra level of unnesting).

    """
    if resource:
        task_label = f"Detecting available codeableConcepts in {resource}..."
    else:
        task_label = "Detecting available codeableConcepts..."  # pragma: no cover
    with base_utils.get_progress_bar() as progress:
        task = progress.add_task(
            task_label,
            # Each column in code_sources requires at most 3 queries to
            # detect valid data is in the DB
            total=len(code_sources),
        )
        for code_source in code_sources:
            code_source.has_data = is_field_populated(
                database=database,
                source_table=code_source.source_table,
                hierarchy=code_source.column_hierarchy,
                expected=code_source.expected,
            )
            progress.advance(task)
    return code_sources


def denormalize_complex_objects(
    database: databases.DatabaseBackend,
    code_sources: list[BaseFHIRResourceConfig],
    resource: str | None = None,
):
    queries = []
    code_sources = _check_data_in_fields(database, code_sources, resource)
    for code_source in code_sources:
        # TODO: This method of pairing classed config objects to
        # specific queries should be considered temporary. This should be
        # replaced at some point by a more generic table schema traversal/
        # generic jinja template approach.
        match code_source:
            case CodeableConceptConfig():
                if code_source.has_data:
                    queries.append(
                        base_templates.get_codeable_concept_denormalize_query(code_source)
                    )
                else:
                    table_cols = [
                        "id",
                        "row",
                        "code",
                        "system",
                        "display",
                        "userSelected",
                    ]
                    col_types = [
                        "varchar",
                        "bigint",
                        "varchar",
                        "varchar",
                        "varchar",
                        "boolean",
                    ]
                    if code_source.extra_fields:
                        table_cols += [f[1] for f in code_source.extra_fields]
                        col_types += ["varchar"] * len(code_source.extra_fields)
                    queries.append(
                        base_templates.get_ctas_empty_query(
                            schema_name=database.schema_name,
                            table_name=code_source.target_table,
                            table_cols=table_cols,
                            table_cols_types=col_types,
                        )
                    )
            case CodingConfig():
                if code_source.has_data:
                    queries.append(base_templates.get_coding_denormalize_query(code_source))
                else:
                    queries.append(
                        base_templates.get_ctas_empty_query(
                            schema_name=database.schema_name,
                            table_name=code_source.target_table,
                            table_cols=["id", "code", "system", "display"],
                        )
                    )

    return queries


def validate_schema(
    database: databases.DatabaseBackend,
    expected_table_cols: dict[str, dict[str, str]],
) -> dict:
    validated_schema = {}
    for table, cols in expected_table_cols.items():
        query = base_templates.get_column_datatype_query(database.schema_name, table, cols.keys())
        try:
            table_schema = database.cursor().execute(query).fetchall()
        except database.operational_errors():
            # A database backend might reasonably raise an exception in cases like
            # the table not existing (Athena does this).
            table_schema = {}
        validated_schema[table] = database.parser().validate_table_schema(cols, table_schema)
    return validated_schema


def is_field_populated(
    *,
    database: databases.DatabaseBackend,
    source_table: str,
    hierarchy: list[tuple],
    expected: list | dict | None = None,
) -> bool:
    """Traverses a complex field and determines if it exists and has data

    Non-core studies that rely on the core tables shouldn't need this method.
    This is just to examine the weird and wonderful world of the raw FHIR tables.

    :keyword database: The database backend
    :keyword source_table: The table to query against
    :keyword hierarchy: a list of tuples defining the FHIR path to the element.
        Each tuple should be of the form ('element_name', dict | list), where
        a dict is a bare nested object and a list is an array object
    :keyword expected: a list of elements that should be present in the field.
        If none, we assume it is a CodeableConcept.
    :returns: a boolean indicating if valid data is present.
    """
    if not is_field_present(
        database=database,
        source_table=source_table,
        source_col=hierarchy[0][0],
        expected=expected,
    ):
        return False
    unnests = []
    source_field = []
    for element in hierarchy:
        if element[1] is list:
            unnests.append(
                {
                    "source_col": ".".join([*source_field, element[0]]),
                    "table_alias": f"{element[0]}_table",
                    "row_alias": f"{element[0]}_row",
                },
            )
            last_table_alias = f"{element[0]}_table"
            last_row_alias = f"{element[0]}_row"
            source_field = [last_table_alias, last_row_alias]
        elif element[1] is dict:
            source_field.append(element[0])
        else:
            raise ValueError(
                f"sql_utils.is_field_populated: Unexpected type {element[1]} for field {element[0]}"
            )
    query = base_templates.get_is_table_not_empty_query(
        source_table=source_table, field=".".join(source_field), unnests=unnests
    )
    res = database.cursor().execute(query).fetchall()
    if len(res) == 0:
        return False
    return True


def is_field_present(
    *,
    database: databases.DatabaseBackend,
    source_table: str,
    source_col: str,
    expected: list | dict | None = None,
) -> bool:
    """Validation check for a column existing, and having the expected schema

    :keyword database: The database backend
    :keyword source_table: The table to query against
    :keyword source_col: The column to check the schema against
    :keyword expected: a list of elements that should be present in source_col.
        If none, we assume it is a CodeableConcept.
    :returns: a boolean indicating if the schema was found.
    """
    if expected is None:
        expected = CODEABLE_CONCEPT

    table_cols = {source_table: {source_col: expected}}
    schema = validate_schema(database, table_cols)

    def _get_all_values(d: dict) -> list:
        all_values = []
        for value in d.values():
            if isinstance(value, dict):
                all_values += _get_all_values(value)
            else:
                all_values.append(value)
        return all_values

    all_schema_values = _get_all_values(schema)
    return all(all_schema_values)



import cumulus_library
from cumulus_library.template_sql import sql_utils


class CoreAllergyIntolerancePrereqBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating AllergyIntolerance codeableConcept tables..."

    def prepare_queries(self, *args, config: cumulus_library.StudyConfig, **kwargs):
        code_sources = [
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("clinicalStatus", dict)],
                target_table="core__allergyintolerance_dn_clinical_status",
                filter_priority=True,
                code_systems=[
                    # Restrict to just this required binding system
                    "http://terminology.hl7.org/CodeSystem/allergyintolerance-clinical",
                ],
            ),
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("verificationStatus", dict)],
                target_table="core__allergyintolerance_dn_verification_status",
                filter_priority=True,
                code_systems=[
                    # Restrict to just this required binding system
                    "http://terminology.hl7.org/CodeSystem/allergyintolerance-verification",
                ],
            ),
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("code", dict)],
                target_table="core__allergyintolerance_dn_code",
            ),
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("reaction", list), ("substance", dict)],
                target_table="core__allergyintolerance_dn_reaction_substance",
                expected={"substance": sql_utils.CODEABLE_CONCEPT},
            ),
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("reaction", list), ("manifestation", list)],
                target_table="core__allergyintolerance_dn_reaction_manifestation",
                expected={"manifestation": sql_utils.CODEABLE_CONCEPT},
            ),
        ]
        self.queries += sql_utils.denormalize_complex_objects(
            config.db, code_sources, "AllergyIntolerance"
        )



"""Database abstraction layer for AWS Athena

See base.py for design rules of thumb - notably, don't use non-PEP249 database
features, like pyathena's async cursors, to simplify cross-db behavior.
"""

import collections
import os
import pathlib
from concurrent import futures

import awswrangler
import boto3
import botocore
import numpy
import pandas
import pyathena
from pyathena.async_cursor import AsyncCursor as AthenaAsyncCursor
from pyathena.common import BaseCursor as AthenaCursor
from pyathena.pandas.cursor import PandasCursor as AthenaPandasCursor
from rich import progress

from cumulus_library import base_utils, errors
from cumulus_library.databases import base, utils


class AthenaDatabaseBackend(base.DatabaseBackend):
    """Database backend that can talk to AWS Athena"""

    connection: None | AthenaCursor

    def __init__(
        self,
        region: str,
        work_group: str,
        profile: str,
        schema_name: str,
        max_concurrent: int | None = None,
    ):
        super().__init__(schema_name)

        self.db_type = "athena"
        self.region = region
        self.work_group = work_group
        self.profile = profile
        self.schema_name = schema_name
        self.connection = None
        self.connect_kwargs = {}
        self.max_concurrent = max_concurrent or 20

    def init_errors(self):  # pragma: no cover
        return ["COLUMN_NOT_FOUND", "TABLE_NOT_FOUND"]

    def connect(self):
        # the profile may not be required, provided the above three AWS env vars
        # are set. If both are present, the env vars take precedence
        if self.profile is not None:
            self.connect_kwargs["profile_name"] = self.profile
        for aws_env_name in [
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
            "AWS_SESSION_TOKEN",
        ]:
            if aws_env_val := os.environ.get(aws_env_name):
                self.connect_kwargs[aws_env_name.lower()] = aws_env_val

        self.connection = pyathena.connect(
            region_name=self.region,
            work_group=self.work_group,
            schema_name=self.schema_name,
            **self.connect_kwargs,
        )

    def cursor(self) -> AthenaCursor:
        return self.connection.cursor()

    def async_cursor(self) -> AthenaAsyncCursor:
        return self.connection.cursor(cursor=AthenaAsyncCursor, max_workers=self.max_concurrent)

    def pandas_cursor(self) -> AthenaPandasCursor:
        return self.connection.cursor(cursor=AthenaPandasCursor)

    def execute_as_pandas(
        self, sql: str, chunksize: int | None = None
    ) -> (pandas.DataFrame | collections.abc.Iterator[pandas.DataFrame], list[tuple]):
        query = self.pandas_cursor().execute(sql, chunksize=chunksize)
        return query.as_pandas(), query.description

    def parser(self) -> base.DatabaseParser:
        return AthenaParser()

    def operational_errors(self) -> tuple[type[Exception], ...]:
        return (pyathena.OperationalError,)

    def col_parquet_types_from_pandas(self, field_types: list) -> list:
        output = []
        for field in field_types:
            match field:
                case numpy.dtypes.ObjectDType():
                    output.append("STRING")
                case pandas.core.arrays.integer.Int64Dtype() | numpy.dtypes.Int64DType():
                    output.append("INT")
                case numpy.dtypes.Float64DType():
                    output.append("DOUBLE")
                case numpy.dtypes.BoolDType():
                    output.append("BOOLEAN")
                case numpy.dtypes.DateTime64DType():
                    output.append("TIMESTAMP")
                case _:
                    raise errors.CumulusLibraryError(
                        f"Unsupported pandas type {type(field)} found."
                    )
        return output

    def upload_file(
        self,
        *,
        file: pathlib.Path,
        study: str,
        topic: str,
        remote_filename: str | None = None,
        force_upload=False,
    ) -> str | None:
        # We'll investigate the connection to get the relevant S3 upload path.
        workgroup = self.connection._client.get_work_group(WorkGroup=self.work_group)
        wg_conf = workgroup["WorkGroup"]["Configuration"]["ResultConfiguration"]
        s3_path = wg_conf["OutputLocation"]
        bucket = "/".join(s3_path.split("/")[2:3])
        key_prefix = "/".join(s3_path.split("/")[3:])
        encryption_type = wg_conf.get("EncryptionConfiguration", {}).get("EncryptionOption")
        if encryption_type is None:
            raise errors.AWSError(
                f"The workgroup {self.work_group} is not reporting an encrypted status. "
                f"Either bucket {bucket} is not encrypted, or the workgroup is overriding the "
                "bucket's default encryption. Please confirm the encryption status of both. "
                "Cumulus requires encryption as part of ensuring safety for limited data sets"
            )
        kms_arn = wg_conf.get("EncryptionConfiguration", {}).get("KmsKey", None)
        s3_key = f"{key_prefix}cumulus_user_uploads/{self.schema_name}/{study}/{topic}"
        if not remote_filename:
            remote_filename = file.name

        session = boto3.Session(profile_name=self.connection.profile_name)
        s3_client = session.client("s3")
        if not force_upload:
            res = s3_client.list_objects_v2(
                Bucket=bucket,
                Prefix=f"{s3_key}/{remote_filename}",
            )
            if res["KeyCount"] > 0:
                return f"s3://{bucket}/{s3_key}"
        with open(file, "rb") as b_file:
            s3_client.put_object(
                Bucket=bucket,
                Key=f"{s3_key}/{remote_filename}",
                Body=b_file,
                ServerSideEncryption="aws:kms",
                SSEKMSKeyId=kms_arn,
            )
        return f"s3://{bucket}/{s3_key}"

    def _clean_bucket_path(self, client, bucket, res):
        for file in res["Contents"]:
            client.delete_object(Bucket=bucket, Key=file["Key"])

    def export_table_as_parquet(
        self, table_name: str, file_name: str, location: pathlib.Path, *args, **kwargs
    ) -> bool:
        session = boto3.session.Session(
            **self.connect_kwargs,
        )
        s3_client = session.client("s3")
        workgroup = self.connection._client.get_work_group(WorkGroup=self.work_group)
        wg_conf = workgroup["WorkGroup"]["Configuration"]["ResultConfiguration"]
        s3_path = wg_conf["OutputLocation"]
        bucket = "/".join(s3_path.split("/")[2:3])
        output_path = location / f"{file_name}"
        s3_path = f"s3://{bucket}/export/{file_name}"
        # Cleanup location in case there was an error of some kind
        res = s3_client.list_objects_v2(Bucket=bucket, Prefix=f"export/{file_name}")
        if "Contents" in res:
            self._clean_bucket_path(s3_client, bucket, res)

        self.connection.cursor().execute(f"""UNLOAD
            (SELECT * FROM {table_name})
            TO '{s3_path}'
            WITH (format='PARQUET', compression='SNAPPY')
            """)  # noqa: S608
        # UNLOAD is not guaranteed to create a single file. AWS Wrangler's read_parquet
        # allows us to ignore that wrinkle
        try:
            df = awswrangler.s3.read_parquet(s3_path, boto3_session=session)
        except awswrangler.exceptions.NoFilesFound:
            return False
        df = df.sort_values(by=list(df.columns), ascending=False, na_position="first")
        df.to_parquet(output_path)
        res = s3_client.list_objects_v2(Bucket=bucket, Prefix=f"export/{file_name}")
        self._clean_bucket_path(s3_client, bucket, res)
        return True

    def parallel_write(
        self,
        queries: list[str],
        verbose: bool,
        progress_bar: progress.Progress,
        task: progress.Task,
    ) -> None:
        def query_completed(f: futures.Future):
            with base_utils.query_console_output(verbose, query, progress_bar, task):
                pass

        async_cursor = self.async_cursor()
        res = []
        for query in queries:
            _, future = async_cursor.execute(query)
            future.add_done_callback(query_completed)
            res.append((query, future))
        utils.handle_concurrent_errors(res, self.db_type)

    def create_schema(self, schema_name) -> None:
        """Creates a new schema object inside the database"""
        glue_client = boto3.client("glue")
        try:
            glue_client.get_database(Name=schema_name)
        except botocore.exceptions.ClientError:
            glue_client.create_database(DatabaseInput={"Name": schema_name})

    def close(self) -> None:
        if self.connection is not None:  # pragma: no cover
            self.connection.close()


class AthenaParser(base.DatabaseParser):
    def _find_type_len(self, row: str) -> int:
        """Finds the end of a type string like row(...) or array(row(...))"""
        # Note that this assumes the string is well formatted.
        depth = 0
        for i in range(len(row)):
            match row[i]:
                case ",":
                    if depth == 0:
                        break
                case "(":
                    depth += 1
                case ")":
                    depth -= 1
        return i

    def _split_row(self, row: str) -> dict[str, str]:
        # Must handle "name type, name row(...), name type, name row(...)"
        result = {}
        # Real athena doesn't add extra spaces, but our unit tests do for
        # readability, so let's strip out whitespace as we parse.
        while row := row.strip():
            name, remainder = row.split(" ", 1)
            type_len = self._find_type_len(remainder)
            result[name] = remainder[0:type_len]
            row = remainder[type_len + 2 :]  # skip comma and space
        return result

    def parse_found_schema(self, schema: dict[str, str]) -> dict:
        # A sample response for table `observation`, column `component`:
        #   array(row(code varchar, display varchar)),
        #             text varchar, id varchar)
        parsed = {}

        for key, value in schema.items():
            # Strip arrays out, they don't affect the shape of our schema.
            while value.startswith("array("):
                value = value.removeprefix("array(")
                value = value.removesuffix(")")

            if value.startswith("row("):
                value = value.removeprefix("row(")
                value = value.removesuffix(")")
                parsed[key] = self.parse_found_schema(self._split_row(value))
            else:
                parsed[key] = {}

        return parsed



"""Class for generating counts tables from templates"""

import pathlib

import rich

from cumulus_library import BaseTableBuilder, errors, study_manifest
from cumulus_library.builders.statistics_templates import counts_templates

# Defined here for easy overriding by tests
DEFAULT_MIN_SUBJECT = 10


class CountsBuilder(BaseTableBuilder):
    """Extends BaseTableBuilder for counts-related use cases"""

    def __init__(
        self, study_prefix: str | None = None, manifest: study_manifest.StudyManifest | None = None
    ):
        super().__init__()
        if manifest:
            self.study_prefix = manifest.get_study_prefix()
        elif study_prefix:
            c = rich.get_console()
            c.print(
                "[yellow]Warning: providing study_prefix to a CountsBuilder is deprecated"
                " and will be removed in a future version"
            )
            self.study_prefix = study_prefix
        else:
            raise errors.CountsBuilderError(
                "CountsBuilder should be initiated with a valid manifest.toml"
            )

    def get_table_name(self, table_name: str, duration=None) -> str:
        """Convenience method for constructing table name

        :param table_name: table name to add after the study prefix
        :param duration: a time period reflecting the table binning strategy
        """
        if duration:
            return f"{self.study_prefix}__{table_name}_{duration}"
        else:
            return f"{self.study_prefix}__{table_name}"

    def get_where_clauses(
        self, clause: list | str | None = None, min_subject: int | None = None
    ) -> list[str]:
        """Convenience method for constructing arbitrary where clauses.

        :param clause: either a string or a list of sql where statements
        :param min_subject: if clause is none, the bin size for a cnt_subject_ref filter
            (deprecated, use count_[fhir_resource](min_subject) instead)
        """
        if min_subject is None:
            min_subject = DEFAULT_MIN_SUBJECT
        if clause is None:
            return [f"cnt_subject_ref >= {min_subject}"]
        elif isinstance(clause, str):
            return [clause]
        elif isinstance(clause, list):
            return clause
        else:
            raise errors.CountsBuilderError(f"get_where_clauses invalid clause {clause}")

    def get_count_query(
        self,
        table_name: str,
        source_table: str,
        table_cols: str | list[list[str] | counts_templates.CountColumn],
        *args,
        min_subject: int | None = None,
        where_clauses: list[str] | None = None,
        filter_resource: bool | None = False,
        primary_id: str = "subject_ref",
        secondary_id: str | None = None,
        alt_secondary_join_id: str | None = None,
        secondary_table: str | None = None,
        secondary_cols: list[str] = [],
        annotation: counts_templates.CountAnnotation | None = None,
        filter_status: bool | None = False,
        filter_cols: list[list[str]] | counts_templates.FilterColumn | None = None,
        **kwargs,
    ) -> str:
        """Generates a counts table using a template

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :keyword fhir_resource: The type of FHIR resource to count (see
            builders/statistics_templates/count_templates.CountableFhirResource)
        :keyword primary_id: the FHIR ID to count by. If not p
        :keyword secondary_table: a table to join to the source table
        :keyword secondary_id: a second id to striate counts by. if secondary_table is
            also present, will be used to join with the source table, assuming that
            the same column exists in both tables
        :keyword alt_secondary_join_id: if present, uses this field instead of secondary_id
            for joining with the source table
        :keyword secondary_cols: the columns to include from the secondary table
        :keyword annotation: A CountsAnnotation object describing a table to use as
            a metadata annotation source
        :keyword filter_status: if true, will create a filter block at the start of your
            query
        :keyword filter_cols: a series of FilterColumns, or list formatted like
            ['column name', ('desired val 1', 'desired val 2'), True/False for including nulls],
            used to add filtering statements to a filter section of a query. If you also
            want these columns in your output, include them in in either table_cols or
            secondary_cols
        """
        if min_subject is None:
            min_subject = DEFAULT_MIN_SUBJECT

        if not table_name or not source_table or not table_cols:
            raise errors.CountsBuilderError(
                f"count_query missing required arguments. output table: {table_name}"
            )

        for key in kwargs:
            if key == "skip_status_filter":  # pragma: no cover
                # Deprecated as of v5.1.0
                rich.print(
                    "[yellow]CountsBuilder deprecation notice[/yellow]: the behavior of "
                    "'skip_status_filter' is now the default behavior and the argument is "
                    "considered deprecated.\n"
                    "Support for it may be removed in a future version."
                )
            elif key == "patient_link":  # pragma: no cover
                # Deprecated as of v6.0.0
                rich.print(
                    "[yellow]CountsBuilder deprecation notice[/yellow]: 'patient_link' has been "
                    "renamed to 'primary_id'. Future versions may remove support for the "
                    "patient_link argument."
                )
        return counts_templates.get_count_query(
            table_name,
            source_table,
            table_cols,
            *args,
            min_subject=min_subject,
            where_clauses=where_clauses,
            primary_id=primary_id,
            secondary_id=secondary_id,
            alt_secondary_join_id=alt_secondary_join_id,
            secondary_table=secondary_table,
            secondary_cols=secondary_cols,
            annotation=annotation,
            filter_status=filter_status,
            filter_cols=filter_cols,
            **kwargs,
        )

    # ----------------------------------------------------------------------
    # The following function all wrap get_count_query as convenience methods.
    # Now that fhir_resource is no longer a param we're concerned with, the
    # utility of these has gone down somewhat, and the filter_status rules
    # may be more specific to the core study than generally.

    # For now, we're going to leave them here, but we may migrate them to
    # the core study at some point.

    def count_allergyintolerance(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
        annotation: counts_templates.CountAnnotation | None = None,
        filter_status: bool | None = False,
    ) -> str:
        """wrapper method for constructing allergyintolerance counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
        :keyword annotation: A CountAnnotation definining an external annotation source
            (default: 10)
        :keyword filter_status: if True, filters the results by what's commonly considered
            'finished' statuses for this resource
        """
        if filter_status:
            extra_kwargs = {
                "filter_cols": [
                    counts_templates.FilterColumn(
                        name="docStatus", values=["final", "amended"], include_nulls=True
                    ),
                    counts_templates.FilterColumn(
                        name="status",
                        values=["current"],
                        include_nulls=False,
                    ),
                ]
            }
        else:  # pragma: no cover
            extra_kwargs = {}
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            primary_id="patient_ref",
            annotation=annotation,
            **extra_kwargs,
        )

    def count_condition(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
        annotation: counts_templates.CountAnnotation | None = None,
    ) -> str:
        """wrapper method for constructing condition counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :keyword annotation: A CountAnnotation definining an external annotation source
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            filter_resource=True,
            annotation=annotation,
            secondary_id="encounter_ref",
        )

    def count_diagnosticreport(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
        annotation: counts_templates.CountAnnotation | None = None,
        **kwargs,
    ) -> str:
        """wrapper method for constructing diagnosticreport counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :keyword annotation: A CountAnnotation definining an external annotation source
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            annotation=annotation,
        )

    def count_documentreference(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
        annotation: counts_templates.CountAnnotation | None = None,
        filter_status: bool | None = False,
        **kwargs,
    ) -> str:
        """wrapper method for constructing documentreference counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :keyword annotation: A CountAnnotation definining an external annotation source
        :keyword filter_status: if True, filters the results by what's commonly considered
            'finished' statuses for this resource
        """

        if filter_status:
            extra_kwargs = {
                "filter_cols": [
                    counts_templates.FilterColumn(
                        name="docStatus", values=["final", "amended"], include_nulls=True
                    ),
                    counts_templates.FilterColumn(
                        name="status",
                        values=["current"],
                        include_nulls=False,
                    ),
                ]
            }
        else:
            extra_kwargs = {}
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            filter_resource=True,
            annotation=annotation,
            filter_status=filter_status,
            primary_id="subject_ref",
            secondary_table="core__encounter",
            secondary_id="documentreference_ref",
            alt_secondary_join_id="encounter_ref",
            secondary_cols=[counts_templates.CountColumn("class_display", "varchar", None)],
            **extra_kwargs,
        )

    def count_encounter(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
        annotation: counts_templates.CountAnnotation | None = None,
        filter_status: bool | None = False,
        **kwargs,
    ) -> str:
        """wrapper method for constructing encounter counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :param min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :param annotation: A CountAnnotation definining an external annotation source
        :param filter_status: Filters encounters by status fields.
            Note: encounters often have cancelled/entered in error statuses.
            If your study is not filtering this out in your cohort selection,
            you can set this flag to do it for you (as long as the field is present)
        """
        if filter_status:
            extra_kwargs = {
                "filter_cols": [
                    counts_templates.FilterColumn(
                        name="status", values=["finished"], include_nulls=False
                    ),
                ]
            }
        else:
            extra_kwargs = {}
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            annotation=annotation,
            filter_status=filter_status,
            primary_id="subject_ref",
            secondary_id="encounter_ref",
            **extra_kwargs,
        )

    def count_medicationrequest(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
        annotation: counts_templates.CountAnnotation | None = None,
    ) -> str:
        """wrapper method for constructing medicationrequests counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :keyword annotation: A CountAnnotation definining an external annotation source
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            annotation=annotation,
        )

    def count_observation(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
        annotation: counts_templates.CountAnnotation | None = None,
        filter_status: bool | None = False,
        **kwargs,
    ) -> str:
        """wrapper method for constructing observation counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :keyword annotation: A CountAnnotation definining an external annotation source
        :keyword filter_status: if True, filters the results by what's commonly considered
            'finished' statuses for this resource
        """
        if filter_status:
            extra_kwargs = {
                "filter_cols": [
                    counts_templates.FilterColumn(
                        name="status", values=["final", "amended"], include_nulls=False
                    ),
                ]
            }
        else:
            extra_kwargs = {}
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            annotation=annotation,
            filter_status=filter_status,
            primary_id="subject_ref",
            secondary_table="core__encounter",
            secondary_id="observation_ref",
            alt_secondary_join_id="encounter_ref",
            secondary_cols=[counts_templates.CountColumn("class_display", "varchar", None)],
            **extra_kwargs,
        )

    def count_patient(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
        annotation: counts_templates.CountAnnotation | None = None,
    ) -> str:
        """wrapper method for constructing patient counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :keyword annotation: A CountAnnotation definining an external annotation source
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            annotation=annotation,
        )

    def count_procedure(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
        annotation: counts_templates.CountAnnotation | None = None,
    ) -> str:
        """wrapper method for constructing procedure counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :keyword annotation: A CountAnnotation definining an external annotation source
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            annotation=annotation,
        )

    # End of wrapper section
    # ----------------------------------------------------------------------

    def write_counts(self, filepath: str):
        """Convenience method for writing counts queries to disk

        :param filepath: path to file to write queries out to.
        """
        self.prepare_queries()
        self.comment_queries()
        self.write_queries(path=pathlib.Path(filepath))

    def prepare_queries(
        self, *args, manifest: study_manifest.StudyManifest | None = None, **kwargs
    ):
        """Prepare count queries

        This should be overridden in any count generator. See studies/core/count_core.py
        for an example
        """
        pass



"""Module for extracting US core extensions from patient records"""

import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "patient": {
        "id": [],
        "gender": [],
        "address": {"postalCode": {}},
        "birthDate": [],
    }
}


class PatientBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Patient tables..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        """constructs queries related to patient extensions of interest

        :param config: A study config object
        """
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(core_templates.get_core_template("patient", validated_schema))



"""Handles the creation of new tables"""

import contextlib
import importlib.util
import inspect
import pathlib
import re
import sys
import tomllib
import zipfile

import rich
import sqlglot
from rich import progress

from cumulus_library import (
    BaseTableBuilder,
    base_utils,
    databases,
    enums,
    errors,
    log_utils,
    study_manifest,
)
from cumulus_library.builders import (
    file_upload_builder,
    protected_table_builder,
    psm_builder,
    valueset_builder,
)

######### public functions #########


def build_study(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    db_parser: databases.DatabaseParser = None,
    continue_from: str | None = None,
    file_list: list | None = None,
    prepare: bool = False,
    data_path: pathlib.Path | None = None,
) -> None:
    """Creates tables in the schema by iterating through the sql_config.file_names

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword db_parser: a parser for the target database
    :keyword continue_from: Name of a file to resume table creation from
    :keyword file_list: If provided, a list of files to build; otherwise pulled from manifest
    :keyword prepare: If true, will render query instead of executing
    :keyword data_path: If prepare is true, the path to write rendered data to
    """
    if prepare:
        _check_if_preparable(manifest.get_study_prefix())
    if file_list is None:
        file_list = manifest.get_file_list(continue_from)
    if prepare:
        data_dir = data_path / manifest.get_study_prefix()
        for file in data_dir.glob("*"):
            if file.is_file():  # pragma: no cover
                file.unlink()

    # a parallel manifest looks like:
    # [file_config.file_names]
    # "step_1" = ["table_1.py",]
    # "step 2" = ["table_2.sql","config.toml"]
    if isinstance(file_list, dict):
        parallel = True

    # a serial manifest looks like this:
    # [file_config]
    # file_names = ["table_1.py","table_2.sql","config.toml"]
    else:
        parallel = False
    for run_stage, entry in enumerate(file_list):
        if isinstance(file_list, dict):
            label = entry
            files = file_list[label]
        else:
            files = [entry]
        query_count = 0
        queries = []
        explicit_serial_queries = []
        for file in files:
            if file.endswith(".py"):
                b_queries, parallel_allowed = _run_builder(
                    config=config,
                    manifest=manifest,
                    filename=file,
                    db_parser=db_parser,
                    data_path=data_path,
                    prepare=prepare,
                    parallel=parallel,
                    query_count=query_count,
                    run_stage=run_stage,
                )
                if parallel_allowed:
                    queries += b_queries
                else:
                    # If you're running in explicit serial mode, you can skip the parallel
                    # collector.
                    # Explicit serial queries are not actually used, but this is kept
                    # here as a debugging convenience if you ever need to look at
                    # the total number of manually executed queries.
                    explicit_serial_queries += b_queries
            elif file.endswith(".toml"):
                w_queries, parallel_allowed = _run_workflow(
                    config=config,
                    manifest=manifest,
                    filename=file,
                    data_path=data_path,
                    prepare=prepare,
                    parallel=parallel,
                    query_count=query_count,
                    run_stage=run_stage,
                )
                if parallel_allowed:
                    queries = queries + w_queries
            elif file.endswith(".sql"):
                queries = queries + _run_raw_queries(
                    config=config,
                    manifest=manifest,
                    filename=file,
                    data_path=data_path,
                    prepare=prepare,
                    parallel=parallel,
                    query_count=query_count,
                    run_stage=run_stage,
                )
            else:
                raise errors.StudyManifestParsingError
            for query in queries:
                _check_query_for_errors(config, manifest, query, file)
        if parallel:
            query_count += len(queries)
            if query_count > 0:
                with base_utils.get_progress_bar() as progress_bar:
                    task = progress_bar.add_task(
                        f"Building {label} tables...",
                        total=query_count,
                        visible=not config.verbose,
                    )
                    config.db.parallel_write(
                        queries=queries,
                        verbose=config.verbose,
                        progress_bar=progress_bar,
                        task=task,
                    )
    if prepare:
        with zipfile.ZipFile(
            f"{data_path}/{manifest.get_study_prefix()}.zip", "w", zipfile.ZIP_DEFLATED
        ) as z:
            for file in data_dir.iterdir():
                z.write(file, file.relative_to(data_dir))


def build_matching_files(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    builder: str | None,
    db_parser: databases.DatabaseParser = None,
    prepare: bool,
    data_path: pathlib.Path,
):
    """targets all table builders matching a target string for running

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword builder: filename of a module implementing a TableBuilder
    :keyword db_parser: an object implementing DatabaseParser for the target database
    :keyword prepare: If true, will render query instead of executing
    :keyword data_path: If prepare is true, the path to write rendered data to
    """
    if prepare:
        _check_if_preparable(manifest.get_study_prefix())  # pragma: no cover
    all_generators = manifest.get_all_generators()
    matches = []
    if not builder:  # pragma: no cover
        matches = all_generators
    else:
        for file in all_generators:
            if file.find(builder) != -1:
                matches.append(file)
    if len(matches) == 0:
        rich.print(f"No builders matching {builder} found - is it in your study manifest?")
        sys.exit()
    build_study(
        config,
        manifest,
        db_parser=db_parser,
        file_list=matches,
        prepare=prepare,
        data_path=data_path,
    )


def run_protected_table_builder(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
) -> None:
    """Creates protected tables for persisting selected data across runs

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    """
    ptb = protected_table_builder.ProtectedTableBuilder()
    ptb.execute_queries(
        config=config,
        manifest=manifest,
    )


######### private helper functions #########


def _check_if_preparable(prefix):
    # This list should include any study which requires interrogating the database to
    # find if data is available to query (outside of a toml-driven workflow),
    # which isn't doable as a distributed query
    if prefix in ["core", "discovery", "data_metrics", "example_nlp"]:
        sys.exit(
            f"Study '{prefix}'' does not support prepare mode. It must be run "
            "directly against a target database."
        )


def _execute_build_queries(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    cursor: databases.DatabaseCursor,
    queries: list,
    filename: str,
    progress: progress.Progress,
    task: progress.TaskID,
) -> None:
    """Handler for executing create table queries and displaying console output.

    :param config: A StudyConfig object
    :param manifest: a StudyManifest object
    :param cursor: A DatabaseCursor object
    :param queries: a list of queries read from files in sql_config.file_names
    :param filename: the filename the queries were sourced from
    :param progress: a rich progress bar renderer
    :param task: a TaskID for a given progress bar
    """
    for query in queries:
        _check_query_for_errors(config, manifest, query, filename)
        try:
            with base_utils.query_console_output(config.verbose, query, progress, task):
                cursor.execute(query)

        except Exception as e:  # pragma: no cover
            _query_error(
                config,
                manifest,
                query,
                filename,
                "You can debug issues with this query using `sqlfluff lint`, "
                "or by executing the query directly against the database.\n"
                f"Error: {e}",
            )


def _render_output(
    study_name: str,
    outputs: list,
    data_path: pathlib.Path,
    filename: str,
    count: int,
    run_stage: int,
    *,
    is_toml: bool = False,
):
    if is_toml:
        suffix = "toml"
    else:
        suffix = "sql"
    for index, output in enumerate(outputs):
        if is_toml:
            name = "config"
        else:
            # This regex attempts to discover the table name, via looking for the first
            # dunder, and then gets the start of the line its on as a non-quote requiring
            # part of a file name. So for example, finding SQL that looks like this:
            #   CREATE TABLE foo__bar AS (varchar baz)
            # would result in `create_table_foo__bar` being assigned to name. The goal
            # is to make this at least mildly parsable at the file system level if someone
            # is reviewing a prepared study
            name = re.search(r"(.*)__\w*", output)[0].lower().replace(" ", "_")
        new_filename = f"{run_stage:04d}.{filename.rsplit('.', 1)[0]}.{index:02d}.{name}.{suffix}"
        file_path = data_path / f"{study_name}/{new_filename}"

        file_path.parent.mkdir(exist_ok=True, parents=True)
        with open(file_path, "w", encoding="UTF-8") as f:
            f.write(output)


@contextlib.contextmanager
def _temporary_sys_path(add: pathlib.Path) -> None:
    orig_path = list(sys.path)
    try:
        sys.path.insert(0, str(add))
        yield
    finally:
        sys.path = orig_path


######### handlers for running different file types #########


def _run_builder(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    filename: str,
    run_stage: int,
    db_parser: databases.DatabaseParser = None,
    write_reference_sql: bool = False,
    doc_str: str | None = None,
    prepare: bool = False,
    parallel: bool = False,
    data_path: pathlib.Path,
    query_count: int | None = None,
) -> tuple[list[str], bool]:
    """Loads a table builder from a file.

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword filename: filename of a module implementing a TableBuilder
    :keyword db_parser: an object implementing DatabaseParser for the target database
    :keyword write_reference_sql: if true, writes sql to disk inside a study's directory
    :keyword doc_str: A string to insert between queries written to disk
    :keyword prepare: If true, will render query instead of executing
    :keyword data_path: If prepare is true, the path to write rendered data to
    :keyword query_count: if prepare is true, the number of queries already rendered
    :returns: a list of queries, a boolean indicating if parallel runs are allowed
    """

    # Since we have to support arbitrary user-defined python files here, we
    # jump through some importlib hoops to import the module directly from
    # a source file defined in the manifest.
    spec = importlib.util.spec_from_file_location(
        "table_builder", f"{manifest._study_path}/{filename}"
    )
    table_builder_module = importlib.util.module_from_spec(spec)
    sys.modules["table_builder"] = table_builder_module
    # Inject the study dir into sys.path so that builders can import
    # from surrounding utility code, even if the study isn't installed.
    # (i.e. you're working from a git checkout and do something like `-s .`)
    with _temporary_sys_path(manifest._study_path.parent):
        spec.loader.exec_module(table_builder_module)

    # We're going to find all subclasses of BaseTableBuild in this file.
    # Since BaseTableBuilder itself is a valid subclass of BaseTableBuilder,
    # we'll detect and skip it. If we don't find any subclasses,
    # we'll punt.
    table_builder_subclasses = []
    for _, cls_obj in inspect.getmembers(table_builder_module, inspect.isclass):
        if issubclass(cls_obj, BaseTableBuilder) and cls_obj != BaseTableBuilder:
            table_builder_subclasses.append(cls_obj)

    if len(table_builder_subclasses) == 0:
        raise errors.StudyManifestParsingError(
            f"Error loading {manifest._study_path}{filename}\n"
            "Custom builders must extend the BaseTableBuilder class."
        )
    # Remove instances of intermediate classes, if present (usually not)
    table_builder_subclasses = list(
        filter(lambda x: x.__name__ != "CountsBuilder", table_builder_subclasses)
    )

    # We'll get the subclass, initialize it, run the executor code, and then
    # remove it so it doesn't interfere with the next python module to
    # execute, since the subclass would otherwise hang around.
    table_builder_class = table_builder_subclasses[0]
    table_builder = table_builder_class(manifest=manifest)
    parallel_allowed = parallel and table_builder.parallel_allowed
    if write_reference_sql:
        prefix = manifest.get_study_prefix()
        table_builder.prepare_queries(config=config, manifest=manifest, parser=db_parser)
        for query_pos in range(len(table_builder.queries)):
            if "s3://" in table_builder.queries[query_pos]:
                table_builder.queries[query_pos] = re.sub(
                    f"s3://(.+){prefix}",
                    f"s3://bucket/db_path/{prefix}",
                    table_builder.queries[query_pos],
                )
        table_builder.comment_queries(doc_str=doc_str)
        new_filename = pathlib.Path(f"{filename}").stem + ".sql"
        table_builder.write_queries(
            path=pathlib.Path(f"{manifest._study_path}/reference_sql/" + new_filename)
        )
    elif prepare:
        table_builder.prepare_queries(
            config=config,
            manifest=manifest,
            parser=db_parser,
        )
        _render_output(
            manifest.get_study_prefix(),
            table_builder.queries,
            data_path,
            filename,
            query_count,
            run_stage,
        )

    elif parallel_allowed:
        table_builder.prepare_queries(
            config=config,
            manifest=manifest,
            parser=db_parser,
        )
    else:
        table_builder.execute_queries(
            config=config,
            manifest=manifest,
            parser=db_parser,
        )
    # After running the executor code, we'll remove
    # it so it doesn't interfere with the next python module to
    # execute, since the subclass would otherwise hang around.
    del sys.modules[table_builder_module.__name__]
    del table_builder_module

    return table_builder.queries, parallel_allowed


def _run_raw_queries(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    filename: str,
    *,
    data_path: pathlib.Path | None,
    prepare: bool,
    parallel: bool = False,
    query_count: int,
    run_stage: int,
) -> list[str]:
    """Creates tables in the schema by iterating through the sql_config.file_names

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :param filename: the name of the sql file to read
    :param prepare: If true, will render query instead of executing
    :param data_path: If prepare is true, the path to write rendered data to
    :param query_count: the number of queries currently processed
    :returns: a list of queries
    """
    raw_queries = []
    cleaned_queries = []
    for query in base_utils.parse_sql(base_utils.load_text(f"{manifest._study_path}/{filename}")):
        raw_queries.append(query)
    for query in raw_queries:
        query = base_utils.update_query_if_schema_specified(query, manifest)
        query = query.replace(
            f"`{manifest.get_study_prefix()}__",
            "`",
        )
        cleaned_queries.append(query)
    if prepare:
        _render_output(
            manifest.get_study_prefix(),
            cleaned_queries,
            data_path,
            filename,
            query_count,
            run_stage,
        )
        return cleaned_queries
    if not parallel:
        # We'll explicitly create a cursor since recreating cursors for each
        # table in a study is slightly slower for some databases
        cursor = config.db.cursor()
        # We want to only show a progress bar if we are :not: printing SQL lines
        with base_utils.get_progress_bar(disable=config.verbose) as progress:
            task = progress.add_task(
                f"Building tables from {filename}...",
                total=len(cleaned_queries),
                visible=not config.verbose,
            )
            _execute_build_queries(
                config=config,
                manifest=manifest,
                cursor=cursor,
                queries=cleaned_queries,
                filename=filename,
                progress=progress,
                task=task,
            )
    return cleaned_queries


def _run_workflow(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    filename: str,
    prepare: str,
    data_path: pathlib.Path,
    query_count: int,
    run_stage: int,
    parallel: bool = False,
) -> tuple[list[str], bool]:
    """Loads workflow config from toml definitions and executes workflow

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :param filename: Filename of the workflow config
    :param prepare: If true, will render query instead of executing
    :param data_path: If prepare is true, the path to write rendered data to
    :param query_count: if prepare is true, the number of queries already rendered
    :returns: a list of queries
    """
    toml_path = pathlib.Path(f"{manifest._study_path}/{filename}")
    if prepare:
        with open(toml_path, encoding="utf-8") as file:
            workflow_config = file.read()
        _render_output(
            manifest.get_study_prefix(),
            [workflow_config],
            data_path,
            filename,
            query_count,
            run_stage,
            is_toml=True,
        )
        return ([], False)

    # This open is a bit redundant with the open inside of the PSM builder,
    # but we're letting it slide so that builders function similarly
    # across the board
    safe_timestamp = base_utils.get_tablename_safe_iso_timestamp()
    with open(toml_path, "rb") as file:
        workflow_config = tomllib.load(file)
        config_type = workflow_config["config_type"]
        target_table = workflow_config.get("target_table", workflow_config.get("table_prefix", ""))

    match config_type:
        case "psm":
            existing_stats = []
            if not config.stats_build:
                existing_stats = (
                    config.db.cursor()
                    .execute(
                        "SELECT view_name FROM "  # noqa: S608
                        f"{manifest.get_study_prefix()}__{enums.ProtectedTables.STATISTICS.value}"
                    )
                    .fetchall()
                )
            if (target_table,) in existing_stats and not config.stats_build:
                return ([], False)
            builder = psm_builder.PsmBuilder(
                toml_config_path=toml_path,
                config=config,
                workflow_config=workflow_config,
                data_path=manifest.data_path / f"{manifest.get_study_prefix()}/psm",
            )
        case "valueset":
            builder = valueset_builder.ValuesetBuilder(
                toml_config_path=toml_path,
                config=config,
                workflow_config=workflow_config,
                data_path=manifest.data_path / f"{manifest.get_study_prefix()}/valueset",
            )
        case "file_upload":
            builder = file_upload_builder.FileUploadBuilder(
                toml_config_path=toml_path,
                config=config,
                workflow_config=workflow_config,
            )
        case _:  # pragma: no cover
            raise errors.StudyManifestParsingError(
                f"{toml_path} references an invalid workflow type {config_type}."
            )
    # we don't have a workflow that will run in parallel mode successfully today,
    # so we'll skip this
    # TODO: revisit after workflow mode added for counts builders
    if parallel and builder.parallel_allowed:  # pragma: no cover
        builder.prepare_queries(
            config=config,
            manifest=manifest,
            table_suffix=safe_timestamp,
        )
    else:
        builder.execute_queries(
            config=config,
            manifest=manifest,
            table_suffix=safe_timestamp,
        )
        if config_type in set(item.value for item in enums.StatisticsTypes):
            log_utils.log_statistics(
                config=config,
                manifest=manifest,
                table_type=config_type,
                table_name=f"{target_table}_{safe_timestamp}",
                view_name=target_table,
            )
    return builder.queries, bool(builder and builder.parallel_allowed)


######### error handlers #########


def _query_error(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    query: str,
    filename: str,
    exit_message: str,
) -> None:
    rich.print(
        "An error occurred executing the following query in ",
        f"{filename}:",
        file=sys.stderr,
    )
    rich.print("--------", file=sys.stderr)
    rich.print(query, file=sys.stderr)
    rich.print("--------", file=sys.stderr)
    log_utils.log_transaction(config=config, manifest=manifest, status=enums.LogStatuses.ERROR)
    rich.print(exit_message)

    if any(init_error in exit_message for init_error in config.db.init_errors()):
        rich.print(
            "Have you initialized your database?\n"
            "https://docs.smarthealthit.org/cumulus/etl/setup/initialization.html"
        )
    sys.exit()


def _check_query_for_errors(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    query: str,
    filename: str,
):
    try:
        table = str(sqlglot.parse_one(query, dialect=config.db.db_type).find(sqlglot.exp.Table))
    except sqlglot.errors.ParseError:
        _query_error(
            config,
            manifest,
            query,
            filename,
            "This query is not a valid SQL query. Check the query against a database "
            "for more debugging information.",
        )
    if f"{manifest.get_study_prefix()}__" not in table and not manifest.get_dedicated_schema():
        _query_error(
            config,
            manifest,
            query,
            filename,
            "This query does not contain the study prefix. All tables should "
            f"start with a string like `{manifest.get_study_prefix()}__`, "
            "and it should be in the first line of the query.",
        )
    if any(
        f"{manifest.get_study_prefix()}__{word.value}_" in table
        for word in enums.ProtectedTableKeywords
    ):
        _query_error(
            config,
            manifest,
            query,
            filename,
            "This query contains a table name which contains a reserved word "
            "immediately after the study prefix. Please rename this table so "
            "that is does not begin with one of these special words "
            "immediately after the double undescore.\n Reserved words: "
            f"{(word.value for word in enums.ProtectedTableKeywords)}",
        )
    if table.count("__") > 1:
        _query_error(
            config,
            manifest,
            query,
            filename,
            "This query contains a table name with more than one '__' in it. "
            "Double underscores are reserved for special use cases. Please "
            "rename this table so the only double undercore is after the "
            f"study prefix, e.g. `{manifest.get_study_prefix()}__`",
        )



import cumulus_library
from cumulus_library.template_sql import sql_utils


class CorePractitionerPrereqBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Practitioner tables..."

    def prepare_queries(self, *args, config: cumulus_library.StudyConfig, **kwargs):
        code_sources = [
            sql_utils.CodeableConceptConfig(
                source_table="practitioner",
                column_hierarchy=[("qualification", list), ("code", dict)],
                target_table="core__practitioner_dn_qualification_code",
                expected={"code": sql_utils.CODEABLE_CONCEPT},
            ),
        ]
        self.queries += sql_utils.denormalize_complex_objects(
            config.db, code_sources, "Practitioner"
        )



import json
import pathlib

import pandas
import rich

from cumulus_library import base_utils, study_manifest
from cumulus_library.apis import umls
from cumulus_library.builders.valueset import valueset_utils


def download_oid_data(
    *,
    steward: str,
    oid: str,
    api_key: str | None = None,
    config: base_utils.StudyConfig | None = None,
    manifest: study_manifest.StudyManifest | None = None,
    force_upload: str | None = None,
    path: pathlib.Path | None = None,
) -> bool:
    """Fetches code definitions (assumed to be RXNorm coded) from VSAC
    :keyword steward: the human readable label for the valueset
    :keyword oid: the ID of the valueset in the VSAC database
    :keyword config: A StudyConfig object. If umls_key is none, will check the
        UMLS_API_KEY env var
    :keyword path: A path to write data to (default: [user cache dir]/[study]/valueset_data)
    :returns: True if file created, false otherwise (mostly for testing)
    """
    if config:
        api_key = config.umls_key
        force_upload = config.force_upload
    path = valueset_utils.get_valueset_cache_dir(path, manifest)
    if not (force_upload) and (path / f"{steward}.parquet").exists():
        rich.print(f"{steward} data present at {path}, skipping download.")
        return False
    rich.print(f"Downloading {steward} to {path}")
    api = umls.UmlsApi(api_key=api_key or api_key)
    output = []

    response = api.get_vsac_valuesets(oid=oid)
    for valueset in response:
        contains = valueset.get("expansion").get("contains", [])
        for data in contains:
            output.append([str(data["system"]), int(data["code"]), str(data["display"])])
    output_df = pandas.DataFrame(output, columns=["SAB", "RXCUI", "STR"])
    output_df.to_csv(path / f"{steward}.tsv", index=False, header=True, sep="\t")
    output_df.to_parquet(path / f"{steward}.parquet")
    with open(path / f"{steward}.json", "w") as f:
        f.write(json.dumps(response))
    return True



"""Error types"""


class CumulusLibraryError(Exception):
    """
    Generic package level error
    """


class AWSError(CumulusLibraryError):
    """Errors from interacting with AWS"""


class ApiError(CumulusLibraryError):
    """Errors from external API calls"""


class CountsBuilderError(CumulusLibraryError):
    """Basic error for CountsBuilder"""


class FileUploadError(CumulusLibraryError):
    """Basic error for FileUploadBuilder"""


class StudyImportError(CumulusLibraryError):
    """Basic error for study importing"""


class StudyManifestFilesystemError(CumulusLibraryError):
    """Errors related to files on disk in StudyManifestParser"""


class StudyManifestParsingError(CumulusLibraryError):
    """Errors related to manifest parsing in StudyManifestParser"""


class StudyManifestQueryError(CumulusLibraryError):
    """Errors related to data queries from StudyManifestParser"""



# Module for generating Propensity Score matching cohorts

import json
import os
import pathlib
import sys
import tomllib
import warnings
from dataclasses import dataclass

import pandas
import rich

from cumulus_library import BaseTableBuilder, StudyManifest, base_utils, databases
from cumulus_library.builders import psmpy_lite
from cumulus_library.builders.statistics_templates import psm_templates
from cumulus_library.template_sql import base_templates


@dataclass
class PsmConfig:
    """Provides expected values for PSM execution

    These values should be read in from a toml configuration file.
    See docs/statistics/propensity-score-matching.md for an example with details about
    the expected values for these fields.

    A word of caution about sampling: the assumptions around PSM analysis require
    that any sampling should not use replacement, so do not turn on panda's dataframe
    replacement. This will mean that very small population sizes (i.e. < 20ish)
    may cause errors to be generated.
    """

    classification_json: str
    pos_source_table: str
    neg_source_table: str
    target_table: str
    primary_ref: str
    count_ref: str
    count_table: str
    dependent_variable: str
    pos_sample_size: int
    neg_sample_size: int
    join_cols_by_table: dict[str, dict]
    seed: int


class PsmBuilder(BaseTableBuilder):
    """TableBuilder for creating PSM tables"""

    display_text = "Building PSM tables..."

    def __init__(self, toml_config_path: str, data_path: pathlib.Path, **kwargs):
        """Loads PSM job details from a PSM configuration file"""
        super().__init__()
        # We're stashing the toml path for error reporting later
        self.toml_path = toml_config_path
        self.data_path = data_path
        try:
            with open(self.toml_path, "rb") as file:
                toml_config = tomllib.load(file)

        except OSError:
            sys.exit(f"PSM configuration not found at {self.toml_path}")
        try:
            toml_dir = pathlib.Path(self.toml_path).parent
            self.config = PsmConfig(
                classification_json=str(toml_dir.joinpath(toml_config["classification_json"])),
                pos_source_table=toml_config["pos_source_table"],
                neg_source_table=toml_config["neg_source_table"],
                target_table=toml_config["target_table"],
                primary_ref=toml_config["primary_ref"],
                dependent_variable=toml_config["dependent_variable"],
                pos_sample_size=toml_config["pos_sample_size"],
                neg_sample_size=toml_config["neg_sample_size"],
                join_cols_by_table=toml_config.get("join_cols_by_table", {}),
                count_ref=toml_config.get("count_ref", None),
                count_table=toml_config.get("count_table", None),
                seed=toml_config.get("seed", 123),
            )
        except KeyError:
            sys.exit(
                f"PSM configuration {toml_config_path} contains missing/invalid keys."
                "Check the PSM documentation for an example config with more details:\n"
                "https://docs.smarthealthit.org/cumulus/library/statistics/propensity-score-matching.html"
            )

    def _get_symptoms_dict(self, path: str) -> dict:
        """convenience function for loading symptoms dictionaries from a json file"""
        with open(path, encoding="UTF-8") as f:
            symptoms = json.load(f)
        return symptoms

    def _get_sampled_ids(
        self,
        cursor: databases.DatabaseCursor,
        schema: str,
        query: str,
        sample_size: int,
        dependent_variable: str,
        is_positive: bool,
    ):
        """Creates a table containing randomly sampled patients for PSM analysis

        To use this, it is assumed you have already identified a cohort of positively
        IDed patients as a manual process.
        :param cursor: A valid DatabaseCusror:
        :param schema: the schema/database name where the data exists
        :param query: a query generated from the psm_dsitinct_ids template
        :param sample_size: the number of records to include in the random sample.
            This should generally be >= 20.
        :param dependent_variable: the name to use for your filtering column
        :param is_positive: defines the value to be used for your filtering column
        """
        df = cursor.execute(query).as_pandas()
        df = df.sort_values(by=[self.config.primary_ref])
        df = (
            df.sample(n=sample_size, random_state=self.config.seed)
            .sort_values(by=[self.config.primary_ref])
            .reset_index()
            .drop("index", axis=1)
        )

        df[dependent_variable] = is_positive
        return df

    def _create_covariate_table(
        self, cursor: databases.DatabaseCursor, schema: str, table_suffix: str
    ):
        """Creates a covariate table from the loaded toml config"""
        # checks for primary & link ref being the same
        source_refs = list({self.config.primary_ref, self.config.count_ref} - {None})
        pos_query = psm_templates.get_distinct_ids(source_refs, self.config.pos_source_table)
        pos = self._get_sampled_ids(
            cursor,
            schema,
            pos_query,
            self.config.pos_sample_size,
            self.config.dependent_variable,
            1,
        )
        neg_query = psm_templates.get_distinct_ids(
            source_refs,
            self.config.neg_source_table,
            join_id=self.config.primary_ref,
            filter_table=self.config.pos_source_table,
        )
        neg = self._get_sampled_ids(
            cursor,
            schema,
            neg_query,
            self.config.neg_sample_size,
            self.config.dependent_variable,
            0,
        )

        cohort = pandas.concat([pos, neg])
        ctas_query = base_templates.get_ctas_query_from_df(
            schema,
            f"{self.config.pos_source_table}_sampled_ids_{table_suffix}",
            cohort,
        )
        self.queries.append(ctas_query)

        dataset_query = psm_templates.get_create_covariate_table(
            target_table=f"{self.config.target_table}_{table_suffix}",
            pos_source_table=self.config.pos_source_table,
            neg_source_table=self.config.neg_source_table,
            table_suffix=table_suffix,
            primary_ref=self.config.primary_ref,
            dependent_variable=self.config.dependent_variable,
            join_cols_by_table=self.config.join_cols_by_table,
            count_ref=self.config.count_ref,
            count_table=self.config.count_table,
        )
        self.queries.append(dataset_query)

    def generate_psm_analysis(
        self,
        cursor: databases.DatabaseCursor,
        manifest: StudyManifest,
        schema: str,
        table_suffix: str,
    ):
        """Runs PSM statistics on generated tables"""
        stats_table = f"{self.config.target_table}_{table_suffix}"
        cursor.execute(base_templates.get_alias_table_query(stats_table, self.config.target_table))
        df = cursor.execute(
            base_templates.get_select_all_query(self.config.target_table)
        ).as_pandas()
        symptoms_dict = self._get_symptoms_dict(self.config.classification_json)
        for dependent_variable, codes in symptoms_dict.items():
            df[dependent_variable] = df["code"].apply(lambda x: 1 if x in codes else 0)
        df = df.drop(columns="code")
        # instance_count present but unused for PSM if table contains a count_ref input
        # (it's intended for manual review)
        df = df.drop(columns="instance_count", errors="ignore")

        columns = []
        if self.config.join_cols_by_table is not None:
            for table_config in self.config.join_cols_by_table.values():
                for column in table_config["included_cols"]:
                    # If there are two elements, it's a SQL column that has been
                    # aliased, so we'll look for the alias name
                    if len(column) == 2:
                        columns.append(column[1])
                    # If there is one element, it's a straight SQL column we can
                    # use with no modification
                    elif len(column) == 1:
                        columns.append(column[0])
                    else:
                        sys.exit(
                            f"PSM config at {self.toml_path} contains an "
                            f"unexpected SQL column definition: {column}."
                            "Check the PSM documentation for valid usages."
                        )

        # This code block is replacing a column which may contain several categories
        # (like male/female/other/unknown for AdministrativeGender), and converts
        # it into a series of 1-hot columns for each distinct value in that column,
        for column in columns:
            encoded_df = pandas.get_dummies(df[column])
            df = pandas.concat([df, encoded_df], axis=1)
            df = df.drop(column, axis=1)
        try:
            psm = psmpy_lite.PsmPy(
                df,
                treatment=self.config.dependent_variable,
                indx=self.config.primary_ref,
                exclude=[],
            )
            # we expect psmpy to autodrop non-matching values, so we'll surpress it
            # mentioning workarounds for this behavior.
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                # This function populates the psm.predicted_data element, which is
                # required for things like the knn_matched() function call
                psm.logistic_ps(balance=True)
                # This function populates the psm.df_matched element
                psm.knn_matched(
                    matcher="propensity_logit",
                    replacement=False,
                    caliper=None,
                    drop_unmatched=True,
                )
                os.makedirs(self.data_path, exist_ok=True)
                histogram = psm.get_histogram().sort_values(by=self.config.primary_ref)
                effect_size = psm.get_effect_size()
                doc_dir = (
                    base_utils.get_user_documents_dir()
                    / f"cumulus-library/{manifest.get_study_prefix()}"
                )
                doc_dir.mkdir(parents=True, exist_ok=True)
                histogram.to_csv(doc_dir / "psm_histogram.csv", index=False)
                effect_size.to_csv(doc_dir / "psm_effect_size.csv", index=False)
                rich.print(f"PSM histogram/effect size data saved to {doc_dir}")
        except ZeroDivisionError:
            sys.exit(
                "Encountered a divide by zero error during statistical graph "
                "generation. Try increasing your sample size."
            )
        except ValueError:
            sys.exit(
                "Encountered a value error during KNN matching. Try increasing your sample size."
            )

    def prepare_queries(
        self,
        config: base_utils.StudyConfig,
        manifest: StudyManifest,
        *args,
        table_suffix: str,
        **kwargs,
    ):
        self._create_covariate_table(config.db.cursor(), config.schema, table_suffix)

    def post_execution(
        self,
        config: base_utils.StudyConfig,
        manifest: StudyManifest,
        *args,
        table_suffix: str | None = None,
        **kwargs,
    ):
        self.generate_psm_analysis(config.db.cursor(), manifest, config.schema, table_suffix)



"""This builder primarily exists to make sure that the FHIR lookup
tables are created before other builders in the core study run, so that
they are available for joins."""

import pathlib

import sqlparse

import cumulus_library


class CorePrereqTableBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating core prerequisite tables..."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parallel_allowed = False

    def prepare_queries(self, *args, **kwargs):
        dir_path = pathlib.Path(__file__).resolve().parents[0]
        prereq_sql = [
            "version.sql",
            "setup.sql",
            "fhir_lookup_tables.sql",
            "fhir_mapping_tables.sql",
        ]
        for sql_file in prereq_sql:
            with open(dir_path / sql_file) as file:
                queries = sqlparse.split(file.read())
                self.queries += queries



import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "documentreference": {
        "id": [],
        "status": [],
        "date": [],
        "docStatus": [],
        "subject": sql_utils.REFERENCE,
        "context": {"encounter": sql_utils.REFERENCE, "period": ["start"]},
        "content": {
            "attachment": {
                "contentType": [],
                "data": [],
                "_data": {"extension": ["url", "valueCode"]},
            },
        },
    }
}


class CoreDocumentreferenceBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating DocumentReference tables..."

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(core_templates.get_core_template("documentreference", validated_schema))



import pathlib

import pandas
import rich
from rich.progress import track

from cumulus_library import base_utils, study_manifest
from cumulus_library.template_sql import base_templates

# Database exporting functions


def reset_counts_exports(
    manifest: study_manifest.StudyManifest,
) -> None:
    """
    Removes exports associated with this study from the ../data_export directory.
    """
    path = pathlib.Path(f"{manifest.data_path}/{manifest.get_study_prefix()}")
    if path.exists():
        # we're just going to remove the count exports - stats exports in
        # subdirectories are left alone by this call
        for file in path.glob("*.*"):
            file.unlink()


def export_study(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    data_path: pathlib.Path,
    archive: bool,
    chunksize: int = 1000000,
) -> None:
    """Exports csvs/parquet extracts of tables listed in export_list
    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword data_path: the path to the place on disk to save data
    :keyword archive: If true, get all study data and zip with timestamp
    :keyword chunksize: number of rows to export in a single transaction
    """
    skipped_tables = []
    reset_counts_exports(manifest)
    if manifest.get_dedicated_schema():
        prefix = f"{manifest.get_dedicated_schema()}."
    else:
        prefix = f"{manifest.get_study_prefix()}__"
    if archive:
        table_query = base_templates.get_show_tables(config.schema, prefix)
        result = config.db.cursor().execute(table_query).fetchall()
        table_list = manifest.get_export_table_list()
        for row in result:
            if row[0] not in table_list:
                table_list.append(study_manifest.ManifestExport(name=row[0], export_type="archive"))
    else:
        table_list = manifest.get_export_table_list()
    path = pathlib.Path(f"{data_path}/{manifest.get_study_prefix()}/")
    path.mkdir(parents=True, exist_ok=True)
    for table in track(
        table_list,
        description=f"Exporting {manifest.get_study_prefix()} data...",
    ):
        table.name = base_utils.update_query_if_schema_specified(table.name, manifest)
        file_name = f"{table.name}.{table.export_type}.parquet"
        if config.db.export_table_as_parquet(table.name, file_name, path):
            parquet_path = path / file_name

            df = pandas.read_parquet(parquet_path)
            df = df.sort_values(
                by=list(df.columns), ascending=False, ignore_index=True, na_position="first"
            )
            df.to_parquet(parquet_path)
            df.to_csv(
                (parquet_path).with_suffix(".csv"),
                index=False,
            )
        else:
            skipped_tables.append(table.name)

    if len(skipped_tables) > 0:
        rich.print("The following tables were empty and were not exported:")
        for table in skipped_tables:
            rich.print(f"  - {table}")
    manifest.write_manifest(data_path)
    base_utils.zip_dir(path, data_path, manifest.get_study_prefix(), archive_csvs=archive)



import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "practitioner": {
        "id": [],
        "active": [],
    }
}


class CorePractitionerBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating Practitioner tables..."

    def prepare_queries(self, *args, config: cumulus_library.StudyConfig, **kwargs):
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(core_templates.get_core_template("practitioner", validated_schema))
